import unittest
import numpy as np
import pandas as pd
from datetime import datetime, timedelta

from preprocessing.data_type_detector import detect_columns_types_summary
from preprocessing.data_transformer import encode_categorical_features

# create dataframe for testing the preprocessing functions:
integer_array = np.random.randint(2, size=(100, 2))
for catagories_numbers in range(5, 50, 10):
    integer_array = np.append(
        integer_array, np.random.randint(catagories_numbers, size=(100, 2)), axis=1
    )
integer_columns = [f"int_col_{x}" for x in range(integer_array.shape[1])]

continuous_array = np.random.randn(100, 10)
continuous_columns = [f"cont_col_{x}" for x in range(continuous_array.shape[1])]

string_array = [
    "pizza",
    "ball",
    "star3",
    "car",
    "01515",
    "cat75",
    "moon",
    "lol",
    "dddd",
    "wtf",
    "xXx",
    "82008",
    "mailbox",
]
string_arrays = np.random.choice(string_array, (100, 10))
string_columns = [f"string_col_{x}" for x in range(string_arrays.shape[1])]

integer_dataframe = pd.DataFrame(integer_array, columns=integer_columns)
continuous_dataframe = pd.DataFrame(continuous_array, columns=continuous_columns)
string_dataframe = pd.DataFrame(string_arrays, columns=string_columns)

date_array = np.array([datetime.now() - timedelta(days=x) for x in range(100)]).reshape(
    100, 1
)
date_dataframe = pd.DataFrame(date_array, columns=["date_col"])
# print(date_dataframe.dtypes)
train_dataframe = pd.concat(
    [date_dataframe, integer_dataframe, continuous_dataframe, string_dataframe], axis=1
)

dataframe_dict = {
    "train": train_dataframe,
    "valid": train_dataframe,
    "test": train_dataframe,
}
columns_set = detect_columns_types_summary(dataframe_dict)
dataframe_dict_encoded = encode_categorical_features(dataframe_dict, string_columns)


class MyTestCase(unittest.TestCase):
    def test_train_dataset(self):
        self.assertEqual(columns_set["train"]["categorical_integer"], integer_columns)
        self.assertEqual(columns_set["train"]["continuous"], continuous_columns)
        self.assertEqual(columns_set["train"]["categorical_string"], string_columns)
        self.assertEqual(columns_set["train"]["date"], ["date_col"])

    def test_valid_dataset(self):
        self.assertEqual(columns_set["valid"]["categorical_integer"], integer_columns)
        self.assertEqual(columns_set["valid"]["continuous"], continuous_columns)
        self.assertEqual(columns_set["valid"]["categorical_string"], string_columns)
        self.assertEqual(columns_set["valid"]["date"], ["date_col"])

    def test_test_dataset(self):
        self.assertEqual(columns_set["test"]["categorical_integer"], integer_columns)
        self.assertEqual(columns_set["test"]["continuous"], continuous_columns)
        self.assertEqual(columns_set["test"]["categorical_string"], string_columns)
        self.assertEqual(columns_set["test"]["date"], ["date_col"])

    def test_encoding(self):
        for column_i in string_columns:
            self.assertEqual(
                len(dataframe_dict_encoded["train"][column_i].value_counts()),
                len(string_array),
            )


if __name__ == "__main__":
    unittest.main()



def unify_dataframes(dataframes_dict: dict, _reference: str, ignore_columns: list) -> dict:
    """ Dataframes unifier

    This function ensures that all datasets have the same features after dropping highly correlated features or columns that have constant values.

    :param dict dataframes_dict: A dictionary that contains Pandas dataframes with nested JSON data type e.g.
                dataframes_dict={ "train": train_dataframe, "test": test_dataframe}
    :param str _reference: The name of the Pandas dataframe that will be used as a reference to adjust the features of other dataframes. Usually it is the train dataframe
    :param list ignore_columns: It contains the columns that should be
                ignored when apply scaling e.g. the id and the target.
    :return:
            - dataframes_dict - A dictionary that contains Pandas dataframes.
    """

    remaining_columns = dataframes_dict[_reference].columns

    for key in dataframes_dict.keys():
        if key != _reference:
            dataframe = dataframes_dict[key]
            try:
                dataframes_dict[key] = dataframe[remaining_columns]
            except Exception as e:
                print(f"The set of remaining columns should be modified. Error: {e}")

                modified_remaining_columns = [x for x in remaining_columns if x not in ignore_columns]
                dataframes_dict[key] = dataframe[modified_remaining_columns]
    return dataframes_dict



import sklearn
from sklearn import *
from sklearn.model_selection import GridSearchCV, StratifiedKFold, KFold


def train_sklearn_grid_search(x_train, y_train, model_type, hyperparameters):
    """ Sklearn Model with GridSearch

    Train a sklearn classification/regression model by picking the best values from a parameter grid
    using sklearn.model_selection.GridSearchCV.

    :param np.array x_train: data features
    :param np.array y_train: data target variable
    :param str model_type: full model name from sklearn library. Examples:
            'sklearn.ensemble.AdaBoostClassifier', 'sklearn.linear_model.LogisticRegression'
    :param dict hyperparameters: contains the following information:\n
            - `dict params_grid`: grid (of not yet determined parameters) used from `GridSearchCV` instance\n
            - `dict params_fixed`: fixed parameters of the model\n
            - `dict params_cv`: parameters for cross validation applied to x_train, y_train\n
            - `objective`: `regression` or  `classification`\n
            - `grid_search_scoring`: scoring functions for grid search
    :return: estimator
    """

    names = model_type.split('.')
    model = sklearn
    for name in names[1:]:
        model = getattr(model, name)
    model = model(**hyperparameters.get('params_fixed', {}))

    if hyperparameters["objective"] == "classification":
        cv = StratifiedKFold(**hyperparameters.get('params_cv', {}))
    else:
        cv = KFold(**hyperparameters.get('params_cv', {}))

    gs_rf = GridSearchCV(model,
                         hyperparameters['params_grid'],
                         scoring=hyperparameters["grid_search_scoring"],
                         cv=cv,
                         n_jobs=-1,
                         refit=hyperparameters["grid_search_scoring"][0])

    gs_rf.fit(x_train, y_train)

    return gs_rf.best_estimator_



import collections
import logging
import os

import numpy as np
import pandas as pd

logger = logging.getLogger(__name__)
formatting = (
    "%(asctime)s: %(levelname)s: File:%(filename)s Function:%(funcName)s Line:%(lineno)d "
    "message:%(message)s"
)
logging.basicConfig(
    filename=os.path.join(os.path.dirname(os.path.dirname(__file__)), "logs/logs.log"),
    level=logging.INFO,
    format=formatting,
)


def _raise_metrics_error():
    raise ValueError("Invalid parameters structure. The metrics should be a list of strings that have maximum two "
                     "elements: r2_score and mean_squared_error for regression, and accuracy_score and"
                     " roc_auc_score for classification")


class StructureValidation:
    """ Parameters structure validation

    Validate the input parameters and raise an Exception if the structure of the parameters is invalid

    :param dict parameters: A dictionary that contains information about the datasets, model type,
                model configurations and training configurations.

    :raise:
            | ValueError:
            | TypeError:
    """

    def __init__(self, parameters):
        """
        :param dict parameters: A dictionary that contains information about the datasets, model type,
                model configurations and training configurations.
        """

        self.parameters = parameters

    def validate_main_keys(self):
        """Main keys validator

        The user has to provide at least those four keys inside the parameters dictionary.

        """

        for key_i in ["data", "split", "model", "metrics"]:
            if key_i not in self.parameters.keys():
                raise ValueError(f"Invalid parameters structure. The key {key_i} is missing")

    def validate_data(self):
        """data key validator

        The following assumptions have to be met:\n
        1. At least there is one dataset inside the data and it should had the key name: train\n
        2. All provided datasets should have two elements: features and target\n
        3. The value of the key features is a pandas dataframe. The target/label should not be between the features\n
        4. The target is a numpy array

        """

        if "train" not in self.parameters["data"].keys():
            raise ValueError("Invalid parameters structure. The key `train` inside `data` is missing")

        for key_i in self.parameters["data"].keys():
            if "features" not in self.parameters["data"][key_i].keys():
                raise ValueError(f"Invalid parameters structure. The key `features` inside the key {key_i} inside "
                                 "`data` is missing")
            if "target" not in self.parameters["data"][key_i].keys():
                raise ValueError(f"Invalid parameters structure. The key `target` inside the key {key_i} inside "
                                 "`data` is missing")
            if not isinstance(self.parameters["data"][key_i]["features"], pd.DataFrame):
                raise TypeError(f"Invalid parameters structure. {key_i}: The features should be a pandas dataframe. "
                                f"The target/labels should not be included in the feature")
            if not isinstance(self.parameters["data"][key_i]["target"], np.ndarray):
                raise TypeError(f"Invalid parameters structure. {key_i}: The target should be a numpy array.")

    def validate_split(self):
        """split key validator

        The following assumptions should be met:\n
        1. There are two elements inside the `split` key: `method` and `split_ratios` or `fold_nr`\n
        2. If the value of the `method` element is `split`, the second element should be `split_ratios`\n
        3. If the value of the `method` element is `kfold`, the second element should be `fold_nr`\n
        4. The `split_ratios` can be either a float or set/list of two floats.\n
        5. The `split_ratios` values should be in ]0, 1[\n
        6. The value of the `fold_nr` should be an integer larger than 1\n
        7. The `method` can take only two values: split` or `kfold`

        """

        if "method" not in self.parameters["split"].keys():
            raise ValueError("Invalid parameters structure. The key `method` inside the key split is missing")

        if self.parameters["split"]["method"] == "split":
            if "split_ratios" not in self.parameters["split"].keys():
                raise ValueError("Invalid parameters structure. The key `split_ratios` inside the key split is missing")

            elif isinstance(self.parameters["split"]["split_ratios"], float) or \
                    isinstance(self.parameters["split"]["split_ratios"], int):
                if self.parameters["split"]["split_ratios"] <= 0 or \
                        self.parameters["split"]["split_ratios"] >= 1:
                    raise ValueError("Invalid parameters structure. split_ratios should be in ]0, 1[")

            elif isinstance(self.parameters["split"]["split_ratios"], collections.Iterable) and \
                    ((len(self.parameters["split"]["split_ratios"]) != 2) |
                     (min(self.parameters["split"]["split_ratios"]) <= 0) |
                     (max(self.parameters["split"]["split_ratios"]) >= 1)):
                raise ValueError("Invalid parameters structure. split_ratios either float or a list or a set that has "
                                 "two floats in the range of ]0, 1[")

        elif self.parameters["split"]["method"] == "kfold":
            if "fold_nr" not in self.parameters["split"].keys():
                raise TypeError("Invalid parameters structure. The key `fold_nr` inside the key split is missing")
            elif not isinstance(self.parameters["split"]["fold_nr"], int) or \
                    self.parameters["split"]["fold_nr"] < 1:
                raise ValueError("Invalid parameters structure. The key `fold_nr` should be an integer larger than 1")

        if self.parameters["split"]["method"] != "split" and self.parameters["split"]["method"] != "kfold":
            raise ValueError("Invalid parameters structure. The split method should be either split or kfold ")

    def validate_model(self):
        """ model key validator

        The following assumptions should be met:\n
        1. The elements `type` and `hyperparameters` should be found inside the values of the key `model`.\n
        2. The type of the value of the `hyperparameters` should be a dictionary.

        """

        if "type" not in self.parameters["model"].keys():
            raise TypeError("Invalid parameters structure. The key type inside the key model is missing")
        elif "hyperparameters" not in self.parameters["model"].keys():
            raise TypeError("Invalid parameters structure. The key hyperparameters inside the key model is missing")
        if not isinstance(self.parameters["model"]["hyperparameters"], dict):
            raise TypeError("Invalid parameters structure. Invalid hyperparameters structure")

    def validate_metrics(self):
        """metrics key validator

        The following assumptions should be met:\n
        1. The value of the `metrics` should be a list.\n
        2. Currently, there are only two regression metrics: r2_score and mean_squared_error, and two classification metrics: accuracy_score and roc_auc_score

        """

        if isinstance(self.parameters["metrics"], list):
            if len(self.parameters["metrics"]) > 2 or len(self.parameters["metrics"]) == 0:
                _raise_metrics_error()

            else:
                for metric_i in self.parameters["metrics"]:
                    if metric_i not in ["r2_score",
                                        "mean_squared_error",
                                        "accuracy_score",
                                        "roc_auc_score"]:
                        _raise_metrics_error()

        else:
            _raise_metrics_error()

    def validate_predict(self):
        """predict key validator

        The following assumptions should be met:\n
        1. If the predict key exists, all of the datasets should have the key `features`\n
        2. The value of the `features` is a pandas dataframe\n
        3. The datasets inside the key `predict` have no target or labels. It is required to predict the target for
        those datasets.

        """

        if "predict" in self.parameters.keys():
            for key_i in self.parameters["predict"]:
                if "features" not in self.parameters["predict"][key_i].keys():
                    raise ValueError(f"Invalid parameters structure. No features were defined for the dataset {key_i}")
                if not isinstance(self.parameters["predict"][key_i]["features"], pd.DataFrame):
                    raise TypeError(f"Invalid parameters structure. {key_i}: The target should be a pandas dataframe.")

    def features_validator(self):
        """Features validator

        All datasets given inside the parameters object should have the same features.

        :return:
        """

        # list of features inside the train dataset
        train_features = list(self.parameters["data"]["train"]["features"].columns)
        for key_i in self.parameters["data"].keys():
            if train_features != list(self.parameters["data"][key_i]["features"].columns):
                raise ValueError(
                    f"Invalid parameters structure. The features of the given dataset data[{key_i}] are not "
                    f"identical with the features of the train dataset")

        if "predict" in self.parameters.keys():
            for key_i in self.parameters["predict"].keys():
                if train_features != list(self.parameters["predict"][key_i]["features"].columns):
                    raise ValueError(
                        f"Invalid parameters structure. The features of the given dataset predict[{key_i}] are not "
                        f"identical with the features of the train dataset")


def parameters_validator(parameters):
    """ Parameters structure validator

    Apply all validation methods defined inside the class `StructureValidation`

    :param dict parameters: A dictionary that contains information about the datasets, model type,
                model configurations and training configurations.

    """

    validator = StructureValidation(parameters)
    validator.validate_main_keys()
    logger.info("All main keys are there")

    validator.validate_data()
    logger.info("Validating data is finished")

    validator.validate_split()
    logger.info("Validating the split data is finished")

    validator.validate_metrics()
    logger.info("Validating the metrics is finished")

    validator.validate_model()
    logger.info("Validating the model data is finished")

    validator.validate_predict()
    logger.info("Validating the predict is finished")

    validator.features_validator()
    logger.info("Validating the features is finished")

    return True



from training.validator import parameters_validator
import numpy as np
import pandas as pd


# create dataframe for testing the preprocessing functions:
def mock_data(number_of_samples):
    integer_array = np.random.randint(2, size=(number_of_samples, 2))
    for categories_numbers in range(5, 50, 10):
        integer_array = np.append(
            integer_array, np.random.randint(categories_numbers, size=(number_of_samples, 2)), axis=1
        )
    integer_columns = [f"int_col_{x}" for x in range(integer_array.shape[1])]

    continuous_array = np.random.randn(number_of_samples, 10)
    continuous_columns = [f"cont_col_{x}" for x in range(continuous_array.shape[1])]

    integer_dataframe = pd.DataFrame(integer_array, columns=integer_columns)
    continuous_dataframe = pd.DataFrame(continuous_array, columns=continuous_columns)

    dataframe = pd.concat([integer_dataframe, continuous_dataframe], axis=1)
    target = (np.sum(continuous_array, axis=1) - 1) / (1 + np.sum(integer_array, axis=1))
    return dataframe, target


train_dataframe, train_target = mock_data(1000)
valid_dataframe, valid_target = mock_data(100)
test_dataframe, test_target = mock_data(100)

# This should not raise any exception
parameters1 = {
    "data": {
        "train": {"features": train_dataframe, "target": train_target},
        "valid": {"features": valid_dataframe, "target": valid_target},
        "test": {"features": test_dataframe, "target": test_target},
    },
    "split": {
        "method": "split",  # "method":"kfold"
        "split_ratios": 0.2,  # foldnr:5 , "split_ratios": 0.8 # "split_ratios":(0.7,0.2)
    },
    "model": {"type": "Ridge linear regression",
              "hyperparameters": {"alpha": 1,  # alpha:optimize
                                  },
              },
    "metrics": ["r2_score", "mean_squared_error"],
    "predict": {
        "test": {"features": test_dataframe}
    }
}

# This should not raise any exception
parameters2 = {
    "data": {
        "train": {"features": train_dataframe, "target": train_target},
        "valid": {"features": valid_dataframe, "target": valid_target},
        "test": {"features": test_dataframe, "target": test_target},
    },
    "split": {
        "method": "split",  # "method":"kfold"
        "split_ratios": 0.2,  # foldnr:5 , "split_ratios": 0.8 # "split_ratios":(0.7,0.2)
    },
    "model": {"type": "Ridge linear regression",
              "hyperparameters": {"alpha": 1,  # alpha:optimize
                                  },
              },
    "metrics": ["accuracy_score", "mean_squared_error"],
    "predict": {
        "test": {"features": test_dataframe}
    }
}

# if the name of the metrics is written wrong, this should raise a value error.
parameters3 = {
    "data": {
        "train": {"features": train_dataframe, "target": train_target},
        "valid": {"features": valid_dataframe, "target": valid_target},
        "test": {"features": test_dataframe, "target": test_target},
    },
    "split": {
        "method": "split",  # "method":"kfold"
        "split_ratios": 0.2,  # foldnr:5 , "split_ratios": 0.8 # "split_ratios":(0.7,0.2)
    },
    "model": {"type": "Ridge linear regression",
              "hyperparameters": {"alpha": 1,  # alpha:optimize
                                  },
              },
    "metrics": ["accuracy_scores", "roc_auc_score"],
    "predict": {
        "test": {"features": test_dataframe}
    }
}

# If the features inside the data.train is not pandas dataframe, this should raise a TypeError
parameters4 = {
    "data": {
        "train": {"features": train_dataframe.values, "target": train_target},
        "valid": {"features": valid_dataframe, "target": valid_target},
        "test": {"features": test_dataframe, "target": test_target},
    },
    "split": {
        "method": "split",  # "method":"kfold"
        "split_ratios": 0.2,  # foldnr:5 , "split_ratios": 0.8 # "split_ratios":(0.7,0.2)
    },
    "model": {"type": "Ridge linear regression",
              "hyperparameters": {"alpha": 1,  # alpha:optimize
                                  },
              },
    "metrics": ["r2_score", "mean_squared_error"],
    "predict": {
        "test": {"features": test_dataframe}
    }
}

# If the predict key exists and the features inside the predict.test is not pandas dataframe,
# this should raise a TypeError
parameters5 = {
    "data": {
        "train": {"features": train_dataframe, "target": train_target},
        "valid": {"features": valid_dataframe, "target": valid_target},
        "test": {"features": test_dataframe, "target": test_target},
    },
    "split": {
        "method": "split",  # "method":"kfold"
        "split_ratios": 0.2,  # foldnr:5 , "split_ratios": 0.2 # "split_ratios":(0.3,0.2)
    },
    "model": {"type": "Ridge linear regression",
              "hyperparameters": {"alpha": 1,  # alpha:optimize
                                  },
              },
    "metrics": ["r2_score", "mean_squared_error"],
    "predict": {
        "test": {"features": test_dataframe.values}
    }
}

# If the target inside the data.train is not a numpy, this should raise a TypeError
parameters6 = {
    "data": {
        "train": {"features": train_dataframe, "target": pd.DataFrame(train_target)},
        "valid": {"features": valid_dataframe, "target": valid_target},
        "test": {"features": test_dataframe, "target": test_target},
    },
    "split": {
        "method": "split",  # "method":"kfold"
        "split_ratios": 0.2,  # foldnr:5 , "split_ratios": 0.2 # "split_ratios":(0.3,0.2)
    },
    "model": {"type": "Ridge linear regression",
              "hyperparameters": {"alpha": 1,  # alpha:optimize
                                  },
              },
    "metrics": ["r2_score", "mean_squared_error"],
    "predict": {
        "test": {"features": test_dataframe}
    }
}

# If the predict key is missing, this should not be a problem
parameters7 = {
    "data": {
        "train": {"features": train_dataframe, "target": train_target},
        "valid": {"features": valid_dataframe, "target": valid_target},
        "test": {"features": test_dataframe, "target": test_target},
    },
    "split": {
        "method": "split",  # "method":"kfold"
        "split_ratios": 0.2,  # foldnr:5 , "split_ratios": 0.2 # "split_ratios":(0.3,0.2)
    },
    "model": {"type": "Ridge linear regression",
              "hyperparameters": {"alpha": 1,  # alpha:optimize
                                  },
              },
    "metrics": ["r2_score", "mean_squared_error"],
}

# If the split_ratios inside split key is not in ]0,1[, a ValueError should be raised
parameters8 = {
    "data": {
        "train": {"features": train_dataframe, "target": train_target},
        "valid": {"features": valid_dataframe, "target": valid_target},
        "test": {"features": test_dataframe, "target": test_target},
    },
    "split": {
        "method": "split",  # "method":"kfold"
        "split_ratios": 1.5,  # foldnr:5 , "split_ratios": 0.2 # "split_ratios":(0.3,0.2)
    },
    "model": {"type": "Ridge linear regression",
              "hyperparameters": {"alpha": 1,  # alpha:optimize
                                  },
              },
    "metrics": ["r2_score", "mean_squared_error"],
    "predict": {
        "test": {"features": test_dataframe}
    }
}

# If the split_ratios inside split key is not in ]0,1[, a ValueError should be raised
parameters9 = {
    "data": {
        "train": {"features": train_dataframe, "target": train_target},
        "valid": {"features": valid_dataframe, "target": valid_target},
        "test": {"features": test_dataframe, "target": test_target},
    },
    "split": {
        "method": "split",  # "method":"kfold"
        "split_ratios": 0,  # foldnr:5 , "split_ratios": 0.2 # "split_ratios":(0.3,0.2)
    },
    "model": {"type": "Ridge linear regression",
              "hyperparameters": {"alpha": 1,  # alpha:optimize
                                  },
              },
    "metrics": ["r2_score", "mean_squared_error"],
    "predict": {
        "test": {"features": test_dataframe}
    }
}

# If the split_ratios inside split key is not in ]0,1[, a ValueError should be raised
parameters10 = {
    "data": {
        "train": {"features": train_dataframe, "target": train_target},
        "valid": {"features": valid_dataframe, "target": valid_target},
        "test": {"features": test_dataframe, "target": test_target},
    },
    "split": {
        "method": "split",  # "method":"kfold"
        "split_ratios": (0.1, 5),  # foldnr:5 , "split_ratios": 0.2 # "split_ratios":(0.3,0.2)
    },
    "model": {"type": "Ridge linear regression",
              "hyperparameters": {"alpha": 1,  # alpha:optimize
                                  },
              },
    "metrics": ["r2_score", "mean_squared_error"],
    "predict": {
        "test": {"features": test_dataframe}
    }
}

# If the split_ratios inside split key is not in ]0,1[, a ValueError should be raised
parameters11 = {
    "data": {
        "train": {"features": train_dataframe, "target": train_target},
        "valid": {"features": valid_dataframe, "target": valid_target},
        "test": {"features": test_dataframe, "target": test_target},
    },
    "split": {
        "method": "split",  # "method":"kfold"
        "split_ratios": (5, 0.5),  # foldnr:5 , "split_ratios": 0.2 # "split_ratios":(0.3,0.2)
    },
    "model": {"type": "Ridge linear regression",
              "hyperparameters": {"alpha": 1,  # alpha:optimize
                                  },
              },
    "metrics": ["r2_score", "mean_squared_error"],
    "predict": {
        "test": {"features": test_dataframe}
    }
}

# If one of the data features or target are missing, it should raise a TypeError
parameters12 = {
    "data": {
        "train": {"target": train_target},
        "valid": {"features": valid_dataframe, "target": valid_target},
        "test": {"features": test_dataframe, "target": test_target},
    },
    "split": {
        "method": "split",  # "method":"kfold"
        "split_ratios": (0.5, 0.5),  # foldnr:5 , "split_ratios": 0.2 # "split_ratios":(0.3,0.2)
    },
    "model": {"type": "Ridge linear regression",
              "hyperparameters": {"alpha": 1,  # alpha:optimize
                                  },
              },
    "metrics": ["r2_score", "mean_squared_error"],
    "predict": {
        "test": {"features": test_dataframe}
    }
}

# If the split method is kfold, we should have foldnr as the second key. Otherwise it raises a TypeError
parameters13 = {
    "data": {
        "train": {"features": train_dataframe, "target": train_target},
        "valid": {"features": valid_dataframe, "target": valid_target},
        "test": {"features": test_dataframe, "target": test_target},
    },
    "split": {
        "method": "kfold",  # "method":"kfold"
        "split_ratios": (0.5, 0.5),  # foldnr:5 , "split_ratios": 0.2 # "split_ratios":(0.3,0.2)
    },
    "model": {"type": "Ridge linear regression",
              "hyperparameters": {"alpha": 1,  # alpha:optimize
                                  },
              },
    "metrics": ["r2_score", "mean_squared_error"],
    "predict": {
        "test": {"features": test_dataframe}
    }
}

# If the split method is kfold, foldnr should be integer >=1. Otherwise, ValueError should be raised
parameters14 = {
    "data": {
        "train": {"features": train_dataframe, "target": train_target},
        "valid": {"features": valid_dataframe, "target": valid_target},
        "test": {"features": test_dataframe, "target": test_target},
    },
    "split": {
        "method": "kfold",  # "method":"kfold"
        "fold_nr": 0.1,  # foldnr:5 , "split_ratios": 0.2 # "split_ratios":(0.3,0.2)
    },
    "model": {"type": "Ridge linear regression",
              "hyperparameters": {"alpha": 1,  # alpha:optimize
                                  },
              },
    "metrics": ["r2_score", "mean_squared_error"],
    "predict": {
        "test": {"features": test_dataframe}
    }
}

# If the type of the model should not be missing. Otherwise, TypeError should be raised
parameters15 = {
    "data": {
        "train": {"features": train_dataframe, "target": train_target},
        "valid": {"features": valid_dataframe, "target": valid_target},
        "test": {"features": test_dataframe, "target": test_target},
    },
    "split": {
        "method": "kfold",  # "method":"kfold"
        "fold_nr": 5,  # foldnr:5 , "split_ratios": 0.2 # "split_ratios":(0.3,0.2)
    },
    "model": {
        "hyperparameters": {"alpha": 1,  # alpha:optimize
                            },
    },
    "metrics": ["r2_score", "mean_squared_error"],
    "predict": {
        "test": {"features": test_dataframe}
    }
}

# If hyperparameters should be a dict
parameters16 = {
    "data": {
        "train": {"features": train_dataframe, "target": train_target},
        "valid": {"features": valid_dataframe, "target": valid_target},
        "test": {"features": test_dataframe, "target": test_target},
    },
    "split": {
        "method": "kfold",  # "method":"kfold"
        "fold_nr": 5,  # foldnr:5 , "split_ratios": 0.2 # "split_ratios":(0.3,0.2)
    },
    "model": {"type": "Ridge linear regression",
              "hyperparameters": [1,  # alpha:optimize
                                  ],
              },
    "metrics": ["r2_score", "mean_squared_error"],
    "predict": {
        "test": {"features": test_dataframe}
    }
}

# positive tests
assert parameters_validator(parameters1)
assert parameters_validator(parameters2)
assert parameters_validator(parameters7)

# negative tests
try:
    assert not parameters_validator(parameters3)
except ValueError:
    assert True

try:
    assert not parameters_validator(parameters4)
except TypeError:
    assert True

try:
    assert not parameters_validator(parameters5)
except TypeError:
    assert True

try:
    assert not parameters_validator(parameters6)
except TypeError:
    assert True

try:
    assert not parameters_validator(parameters8)
except ValueError:
    assert True

try:
    assert not parameters_validator(parameters9)
except ValueError:
    assert True

try:
    assert not parameters_validator(parameters10)
except ValueError:
    assert True

try:
    assert not parameters_validator(parameters11)
except ValueError:
    assert True

try:
    assert not parameters_validator(parameters12)
except ValueError:
    assert True

try:
    assert not parameters_validator(parameters13)
except TypeError:
    assert True

try:
    assert not parameters_validator(parameters14)
except ValueError:
    assert True

try:
    assert not parameters_validator(parameters15)
except TypeError:
    assert True

try:
    assert not parameters_validator(parameters16)
except TypeError:
    assert True



import logging
import os

import numpy as np
import pandas as pd
import category_encoders as ce
from sklearn.preprocessing import OneHotEncoder

logger = logging.getLogger(__name__)
formatting = (
    "%(asctime)s: %(levelname)s: File:%(filename)s Function:%(funcName)s Line:%(lineno)d "
    "message:%(message)s"
)
logging.basicConfig(
    filename=os.path.join(os.path.dirname(os.path.dirname(__file__)), "logs/logs.log"),
    level=logging.INFO,
    format=formatting,
)


def decode_features_with_appearance_frequency(dataframe_dict: dict, reference: str, categorical_features: list) -> dict:
    """ Frequency-based encoder

    :param dict dataframe_dict: A dictionary that contains the dataframes before applying the encoding e.g.
            dataframes_dict={ 'train': train_dataframe, 'test': 'test_dataframe'}
    :param str reference: The name of the dataframe that will be considered when validating the type of the data
    :param list categorical_features: A list of string that contains the name of the columns or features that contain
            categorical data type.
    :return:
            dataframe_dict: A dictionary that contains the dataframes after applying feature encoding
    :rtype: dict

    """

    _reference_dataframe = dataframe_dict[reference]

    for column_i in categorical_features:
        print(f"decoding the column {column_i} based on the value frequency")
        try:
            encoding = _reference_dataframe.groupby([column_i]).size()

            encoding /= _reference_dataframe.shape[0]

            for key_i, dataframe in dataframe_dict.items():
                dataframe[column_i + '_frequency_encoding'] = dataframe[column_i].map(encoding)
                dataframe_dict[key_i] = dataframe

        except Exception as e:
            print(f"error while decoding {column_i} based on value appearance frequency")
            print("Error: ", e)

    return dataframe_dict


def valid_features_detector(dataframe: pd.DataFrame, categorical_features: list,
                            class_number_range: list) -> list:
    """ Feature validator

    The functions checks if the one-hot encoding method should be applied to the given features.

    :param pd.DataFrame dataframe: A pandas dataframe which contain the dataset
    :param list categorical_features: A list of string that contains the name of the columns or features that contain
            categorical data type.
    :param list class_number_range: It is a list of two elements which define the minimum the and maximum number of
                        the classes (unique value) that a feature should contain in order to apply the one-hot encoding
                         to this feature.
    :return:
            valid_features: A list of the features which the encoding will be applied to.
    :rtype: list
    """

    valid_features = [x for x in categorical_features if len(dataframe[x].value_counts()) in range(
        class_number_range[0], class_number_range[1])]
    return valid_features


def encoding_features(encoding_type: str,
                      dataframes_dict: dict,
                      reference: str,
                      categorical_features: list,
                      ignore_columns: list,
                      class_number_range: list = [0, 50],
                      target_name: str = None,
                      drop_encoded_features: bool = True) -> dict:
    """ One-hot encoder

    The function applies one-hot encoding to the categorical features using the Scikit Learn framework implementation.

    :param str encoding_type: The type of the encoding method that will be applied. For example: one-hot, target \n
            For more information please check the following reference:\n
            https://contrib.scikit-learn.org/categorical-encoding/index.html
    :param dict dataframes_dict: A dictionary that contains the dataframes before applying the encoding e.g.
            dataframes_dict={ 'train': train_dataframe, 'test': 'test_dataframe'}
    :param str reference: The name of the dataframe that will be considered when validating the type of the data
    :param list categorical_features: A list of string that contains the name of the columns or features that contain
            categorical data type.
    :param list class_number_range: A list that contains two integers which refer ot the range of the minimum and the
            maximum number of the labels/classes/ categories. If a number of the categories of the feature is not in
            that defined range, one-hot encoding will be not applied to that feature.
            It can be ignored if the encoding type is not one-hot.
    :param list ignore_columns: list of strings which are the columns names. The encoding will not be applied to
            those columns.
    :param str target_name: The name of the column that contains the labels that should be predicted by the model.
                            If the encoding method doesn't require that target, it can be ignored.
    :param bool drop_encoded_features: If True, the encoded features will be dropped from the dataset

    :return:
            dataframes_dict_encoded: A dictionary that contains the dataframes after applying feature encoding e.g.
            dataframes_dict_encoded={ 'train': train_dataframe, 'test': 'test_dataframe'}

    :rtype: dict
    """

    dataframes_dict_encoded = {}

    considered_features = [x for x in categorical_features if x not in ignore_columns]

    encoder = None

    if encoding_type == "one-hot":
        considered_features = valid_features_detector(dataframes_dict[reference],
                                                      categorical_features,
                                                      class_number_range)

        logging.debug(f"considered_features = {considered_features}")

        encoder = OneHotEncoder(sparse=False, handle_unknown="ignore").fit(
            np.array(dataframes_dict[reference][considered_features]))

    if encoding_type == "target":
        if target_name is None:
            raise ValueError("Please define the target_name. It is the label that should be predicted by the model")

        encoder = ce.target_encoder.TargetEncoder().fit(
            np.array(dataframes_dict[reference][considered_features]),
            dataframes_dict[reference][target_name])

    if encoding_type == "frequency":
        dataframes_dict_encoded = decode_features_with_appearance_frequency(dataframes_dict,
                                                                            reference,
                                                                            considered_features)

    else:

        for key_i, dataframe in dataframes_dict.items():
            encoded_data = pd.DataFrame(encoder.transform(np.array(dataframe[considered_features])))
            columns = [f"col_{encoding_type}_encoding_{x}" for x in range(encoded_data.shape[1])]

            logging.debug(f"the number of the {encoding_type} encoded data columns is {len(columns)}")

            encoded_data.columns = columns

            # Encoded features will be dropped.
            if drop_encoded_features:
                dataframe = dataframe.drop(considered_features, axis=1)

            dataframes_dict_encoded[key_i] = pd.concat([dataframe, encoded_data], axis=1, sort=False)

    return dataframes_dict_encoded



import logging
import os
import pickle

import numpy as np

from training.xgboost_train import xgboost_data_preparation_to_predict

logger = logging.getLogger(__name__)
formatting = (
    "%(asctime)s: %(levelname)s: File:%(filename)s Function:%(funcName)s Line:%(lineno)d "
    "message:%(message)s"
)
logging.basicConfig(
    filename=os.path.join(os.path.dirname(os.path.dirname(__file__)), "logs/logs.log"),
    level=logging.DEBUG,
    format=formatting,
)


def predict_unseen_data(predict: dict, models_nr: list, save_models_dir: str, model_type: str) -> np.array:
    """ Target prediction

    This function predicts unseen data given in the dictionary `predict`

    :param dict predict: A dictionary that contains pandas dataframes which are unseen data that will be used to
            predict the target
    :param list models_nr: A list of indexes that will be used to point to the trained models which were saved
            locally after training.
    :param str save_models_dir: The path where the models will be saved.
    :param str model_type: The type of the model that was used to fit the data e.g. linear or lightgbm

    :return:
            - y_prediction - A numpy array that contains the predicted target of the unseen data.
    """

    y_prediction = 0
    model = None

    for data_key, dataframe in predict.items():
        if model_type == "xgboost":
            dataframe_i = predict[data_key]["features"]
            dataframe_i.columns = [x for x in range(len(list(dataframe_i.columns)))]
            array = xgboost_data_preparation_to_predict(dataframe_i)
        else:
            array = predict[data_key]["features"].values

        # reset the variable
        y_prediction = 0

        for model_i in models_nr:
            try:
                with open(os.path.join(save_models_dir, f"{model_type}_{model_i}.pkl"), "rb") as save_model:
                    model = pickle.load(save_model)
                logger.info("The model is loaded successfully: " + f"{model_type}_{model_i}.pkl")
            except Exception as e:
                logger.error(f"Loading the model failed. Error is {e}")

            y_prediction += model.predict(array)

        # Stacking the results of different models using the mean value
        y_prediction = y_prediction / len(models_nr)

    return y_prediction


def model_prediction(predict: dict, models_nr: list, save_models_dir: str, model_type: str) -> np.array:
    """Predictor

    This function applies the function `predict_unseen_data` to predict the target using the data in the `predict`.

    :param dict predict: A dictionary that contains pandas dataframes which are unseen data that will be used to
            predict the target
    :param list models_nr: A list of indexes that will be used to point to the trained models which were saved
            locally after training.
    :param str save_models_dir: The path where the models will be saved.
    :param str model_type: The type of the model that was used to fit the data e.g. linear or lightgbm

    :return:
            - y_prediction - A numpy array that contains the predicted target of the unseen data.
    """

    # Predict the target of unseen data
    y_prediction = 0

    if isinstance(predict, bool):
        logger.info("Nothing will be predicted")
    else:
        try:
            y_prediction = predict_unseen_data(predict, models_nr, save_models_dir, model_type)
            logger.info("Prediction process is finished!")
        except Exception as e:
            logger.error(f"Prediction failed: The error is {e}")

    return y_prediction



import yaml
from blessings import Terminal
from flow_instructions import FlowInstructions

term = Terminal()

with open("flow_instructions_database.yaml", 'r') as stream:
    try:
        commands = yaml.safe_load(stream)
    except yaml.YAMLError as exc:
        print(exc)
for command_key in commands.keys():
    for command in commands[command_key][0]["guide"]:
        eval(command)



import logging
import os
import itertools
import random
from typing import Union

import numpy as np
import pandas as pd
from sklearn import preprocessing
from IPython.display import display

logger = logging.getLogger(__name__)
formatting = (
    "%(asctime)s: %(levelname)s: File:%(filename)s Function:%(funcName)s Line:%(lineno)d "
    "message:%(message)s"
)
logging.basicConfig(
    filename=os.path.join(os.path.dirname(os.path.dirname(__file__)), "logs/logs.log"),
    level=logging.INFO,
    format=formatting,
)
pd.set_option('display.max_rows', None)


def standard_scale_numeric_features(
        dataframe_dict: dict,
        reference_dataframe_key: str,
        columns_to_normalize: list,
        handle_missing_values: bool = True,
) -> dict:
    """ Feature standardizer

    This function standardizes the datesets passed as pandas dataframes by a dictionary. The reference dataframe will be
    used to calculate the statistical properties (mean and standard deviation) and used to normalize other
    dataframes. After standardizing the dataset, the features in each dataset have 0 mean and 1 standard deviation.

    :param dict dataframe_dict: A dictionary that contains multiple pandas dataframes
    :param str reference_dataframe_key: A string that is used to fit the scaler e.g. "train"
    :param list columns_to_normalize: A list of the columns that should be standardized e.g. [col_1, col_2, ..., col_n]
    :param bool handle_missing_values: boolean if true the missing values will be replaced by the value 0

    :return:
        scaled_dataframe_dict: A dictionary of pandas dataframes where the "columns_to_normalize" are normalized
    :rtype: dict
    """

    scaler = None
    scaled_dataframe_dict = {}

    # For scaling the data, the reference, that will be used to fit the StandardScaler, should be selected.
    reference_dataframe = dataframe_dict[reference_dataframe_key]

    # fitting the StandardScaler
    try:
        scaler = preprocessing.StandardScaler().fit(
            reference_dataframe[columns_to_normalize]
        )
    except Exception as e:
        logger.error(f"Error is: {e}")

    # Scaling data
    for key_i, dataframe in dataframe_dict.items():
        scaled = scaler.transform(dataframe[columns_to_normalize])

        # After scaling, the mean values will be 0. Therefore, the missing values will be replaced by the mean value
        if handle_missing_values:
            logger.info("Missing values will be handled")
            scaled[np.isnan(scaled)] = 0

        dataframe[columns_to_normalize] = scaled
        scaled_dataframe_dict[key_i] = dataframe

    logger.info("Scaling data is finished")
    return scaled_dataframe_dict


def encoding_categorical_feature(dataset_dict: dict, feature_name: str,
                                 print_results: Union[bool, int] = True,
                                 print_counter: int = 0) -> dict:
    """ Single categorical feature string encoder

    This function encodes categorical features. It is possible to use train data alone or all train data, validation
    data and test data. If all datesets are provided (i.e. train, valid and test), they will be concatenated first
    and then encoded.

    :param int print_counter: if print_results is int, print counter control printing data to the conosle based on the print_results value.
    :param Union[bool, int] print_results: If False, no data is printed to the console. If True, all data is printed to the console. If an integer n, only the data for n features is printed to the console.
    :param str feature_name: The name of the feature/column that its values should be encoded.
    :param dict dataset_dict: a dictionary of pandas series (i.e one column) that must contain the train data and
                                optionally contains valid data and test data

    :return:
            dataset_dict_encoded: a dictionary of pandas series (i.e one column) after encoding.
    """

    # Replacing the missing values with a special hash value to avoid having the same class of missing values and
    # non-missing values.
    hash_missing_value = hex(random.getrandbits(128))
    logger.debug(f"The hash for the missing values is {hash_missing_value}")

    # Concatenate datasets if needed

    if isinstance(print_results, bool) and print_results:
        print(f"there are {len(dataset_dict)} datasets provided")
    elif isinstance(print_results, int):
        if print_counter < print_results:
            print(f"there are {len(dataset_dict)} datasets provided")

    # the check here is not a good idea because it check each column alone which is not efficient
    valid_dataset_list = []
    valid_dataset_keys = []

    for key_i, dataseries in dataset_dict.items():
        if dataseries.shape[0] > 0:
            valid_dataset_list.append(dataseries)  # get the dataframes
            valid_dataset_keys.append(key_i)  # get the keys

    if len(valid_dataset_list) > 1:
        x_original = pd.concat(valid_dataset_list, axis=0)
    elif len(valid_dataset_list) == 1:
        x_original = valid_dataset_list[0]
    else:
        raise ValueError("No valid dataset was provided")

    # define the encoder
    label_encoder = preprocessing.LabelEncoder()
    label_encoder.fit(x_original.fillna(hash_missing_value))

    dataset_dict_encoded = {}

    for dataset_key in valid_dataset_keys:
        dataset_dict_encoded[dataset_key] = label_encoder.transform(
            dataset_dict[dataset_key].fillna(hash_missing_value)
        )

        labels_nr = len(list(label_encoder.classes_))
        if isinstance(print_results, bool) and print_results:
            print(f"encoding the feature in the dataset {dataset_key}")
            print(f"the number of classes in {feature_name} feature is: {labels_nr}")

        elif isinstance(print_results, int):
            if print_counter < print_results:
                print(f"encoding the feature in the dataset {dataset_key}")
                print(f"the number of classes in {feature_name} feature is: {labels_nr}")

    logger.info(f"Encoding categorical feature {feature_name} process is finished!")
    return dataset_dict_encoded


def encode_categorical_features(dataframe_dict: dict, columns_list: list,
                                print_results: Union[bool, int] = True) -> dict:
    """ Categorical features string encoder

    This function applies the `encoding_categorical_feature` function to each feature in the `columns_list`.

    :param Union[bool, int] print_results: If False, no data is printed to the console. If True, all data is printed to
        the console. If an integer n, only the data for n features is printed to the console.
    :param dict dataframe_dict: a dictionary of Pandas dataframes.
    :param list columns_list: The list of the names of the columns/features that their values should be encoded.

    :return:
            dataframe_dict: a dictionary of Pandas dataframes after encoding.
    """

    print_counter = 0
    for feature_name in columns_list:
        dataset_dict = {}
        try:
            for key_i, dataframe in dataframe_dict.items():
                dataset_dict[key_i] = dataframe[feature_name].astype(str)

            dataset_dict_encoded = encoding_categorical_feature(dataset_dict, feature_name,
                                                                print_results,
                                                                print_counter)
            for key_i, dataseries in dataset_dict_encoded.items():
                dataframe_dict[key_i][feature_name] = dataseries

            logger.info("Encoding all categorical features process is finished!")

        except Exception as e:
            logger.error(f"The Error: {e}")

        print_counter += 1

    if isinstance(print_results, int) and print_results > 0:
        print("the value is considered integer")

        dfs = [v[[c for c in v.columns if c in columns_list]].nunique().to_frame(name=k)
               for k, v in dataframe_dict.items()]
        result = pd.concat(dfs, axis=1, sort=False)

        dfs = [v[[c for c in v.columns if c in columns_list]].apply(lambda x: list(set(x)), axis=0).to_frame(name=k)
               for k, v in dataframe_dict.items()]

        result_all = pd.concat(dfs, axis=1, sort=False)
        result_all = result_all \
            .apply(lambda x: len(set(itertools.chain(*filter(lambda y: type(y) == list, x)))), axis=1) \
            .to_frame(name='all data sets')

        result = pd.concat([result, result_all], axis=1, sort=False)
        result.index.name = 'string columns'

        print(f"Number of unique elements per string column")

        if print_results > 1:
            pd.set_option('display.max_rows', print_results)
        display(result)
        pd.set_option('display.max_rows', None)

    return dataframe_dict



""" Flows pieces collector

This module contains the Flows class, which is a container of the
methods that construct the flows.

The modules depends on the classes and methods which are writen
in other package in this project.

"""

import json
import logging
import os
from typing import Union, Tuple, List, Dict

import numpy as np
import yaml
from blessings import Terminal

from feature_engineering.feature_generator import encoding_features
from flows.flow_instructions import FlowInstructions
from flows.utils import unify_dataframes
from prediction.model_predictor import model_prediction
from preprocessing.data_clean import drop_corr_columns, drop_const_columns
from preprocessing.data_explorer import explore_data
from preprocessing.data_science_help_functions import detect_id_target_problem
from preprocessing.data_transformer import encode_categorical_features, standard_scale_numeric_features
from preprocessing.data_type_detector import detect_columns_types_summary
from preprocessing.json_preprocessor import flat_json
from preprocessing.utils import check_if_target_columns_are_imbalanced
from preprocessing.utils import read_data
from training.training import model_training
from visualization.visualization import compare_statistics

logger = logging.getLogger(__name__)
formatting = (
    "%(asctime)s: %(levelname)s: File:%(filename)s Function:%(funcName)s Line:%(lineno)d "
    "message:%(message)s"
)
logging.basicConfig(
    filename=os.path.join(os.path.dirname(os.path.dirname(__file__)), "logs", "logs.log"),
    level=logging.DEBUG,
    format=formatting,
)

# For colorful and beautiful formatted print()
term = Terminal()

print(term.bold(term.magenta("Welcome to the Data Science Package. First create an object as follows:")))
print(term.bold(term.magenta("For example, use the code below to import the flow 0:")))

print(term.green_on_black("flow = Flows(0)"))
print(term.bold(term.magenta("You can define the `categorical_threshold` which is the maximum number of categories"
                             " that a categorical feature should have before considering it as continuous"
                             " numeric feature. The default value is 50")))
print(term.bold(term.magenta("For example, use the code below to import the flow 0"
                             " with defining the categorical_threshold as 50")))
print(term.green_on_black("flow = Flows(flow_id=0, categorical_threshold=50)"))


class Flows:
    """
    Flows methods container

    A class which meant to be a container for all flows elements.

    :param int flow_id: An integer which points to the flow that the
            use wants to follow.
    :param int categorical_threshold: The maximum number of categories that a categorical feature should have before
                                      considering it as continuous numeric feature.
    :param object commands: It contains the list of the instructions
            that are loaded from the yaml file
    :param list columns_set: A dictionary that contains the features'
                                names sorted in multiple lists based on the
                                type of the data for each given dataset.

    :methods:
            - `guidance` - Evaluate YAML commands
            - `load_data` - Read CSV data
            - `encode_categorical_feature` - Encode the categorical features by changing the string value to numeric values
            - `scale_data` - Scale the numerical value (Feature Standardization - mean = 0, STD = 1)
            - `one_hot_encoding` - Encode categorical features using one-hot encoding method
            - `training_ridge_linear_model` - Train a model using the regression Scikit Learn Ridge linear model implementation
            - `training_lightgbm` - Train a tree-based regression model using the LightGBM implementation

    """

    def __init__(self, flow_id: int, categorical_threshold: int = 50, kl_div_threshold: float = 0.05):
        """

        :param int flow_id: An integer which points to the flow that the use wants to follow.
        :param int categorical_threshold: The maximum number of categories that a categorical feature should have before
                                      considering it as continuous numeric feature.

        """

        self.columns_set = None
        self.flow_id = flow_id
        self.flow_steps = {}
        self.categorical_threshold = categorical_threshold
        self.kl_div_threshold = kl_div_threshold

        # load the yaml file that contains the instructions of the flow
        flow_path = os.path.join(os.path.dirname(os.path.dirname(__file__)), "flows", f"flow_{self.flow_id}.json")
        flow_instruction_database = os.path.join(os.path.dirname(os.path.dirname(__file__)),
                                                 "flows", "flow_instructions_database.yaml")
        try:
            with open(flow_instruction_database, 'r') as stream:
                self.commands = yaml.safe_load(stream)
                logger.info("flow_instruction_database.yaml was loaded successfully.")
        except yaml.YAMLError as exc:
            logger.error(f"Loading the yaml file failed. The error is {exc}")

        try:
            with open(flow_path, 'r') as stream:
                self.flow_steps = json.load(stream)
                logger.info(f"flow_{self.flow_id}.json was loaded successfully.")
        except Exception as exc:
            logger.error(f"Loading the json file failed. The error is {exc}")

        # First thing first: tell the user how to read the data and explore the features
        FlowInstructions.read_data()
        FlowInstructions.explore_data()

    def guidance(self, step_ext: object):
        """ YAML evaluator

        This function executes the command that is written in the yaml file under a certain step

        :param object step_ext: It can be an integer that points to a certain step e.g. 1 or a combination of both
                an integer and a letter to point to a sub-step e.g. 1_a

        """

        for command in self.commands[step_ext][0]["guide"]:
            try:
                eval(command)
                logger.info(f"The command in the step {step_ext} was executed successfully")
            except Exception as e:
                logger.error(f"Failed executing the command in the step {step_ext}. Error is: {e}")

    def load_data(self, path: str, files_list: list, rows_amount: int = 0) -> Tuple[Dict, List]:
        """ Data reader

        This function reads data from CSV files and returns a dictionary
        that contains Pandas dataframes e.g. dataframes_dict={"train":
        train_dataframe, "test": test_dataframe}

        After reading the data, the function provides a summary of
        each dataset.

        After presenting the summary, the function tries to detect
        which column may contain the ids and which column can be
        the target (labels). Based on the values of the target,
        the function can tell if the problem which should be solved
        is a regression or classification problem.

        :param str path: The path to the data.
        :param list files_list: A list of strings which are the
                names of the files
        :param int rows_amount: The number of rows that should be read from the CSV file. If 0, all rows will be read.

        :return:
                - dataframes_dict - A dictionary that contains Pandas dataframes.
                - columns_set - A dictionary that contains the features' names sorted in multiple lists based on the type of the data for each given dataset.

        :example:

        >>> path = "./data"
        >>> files_list = ['train.csv','test.csv']

        >>> dataframes_dict={"train": train_dataframe, "test": test_dataframe}
        """

        function_id = "0"

        dataframes_dict = read_data(path, files_list, rows_amount)

        self.columns_set = detect_columns_types_summary(dataframes_dict, threshold=self.categorical_threshold)

        _, _, possible_problems = detect_id_target_problem(dataframes_dict)

        _ = check_if_target_columns_are_imbalanced(dataframes_dict, possible_problems, self.kl_div_threshold)

        try:
            self.guidance(self.flow_steps[function_id])
        except Exception as e:
            print(f"It seems that the next step is not defined. Error{e}")

        return dataframes_dict, self.columns_set

    def encode_categorical_feature(self, dataframes_dict: dict, print_results: Union[bool, int] = False,
                                   _reference: Union[bool, str] = False) -> Tuple[Dict, List]:
        """ Categorical features encoder

        This function encodes the categorical features by replacing
        the strings with integers

        :param Union[bool, int] print_results: If False, no data is printed to the console. If True, all data is printed to the console. If an integer n, only the data for n features is printed to the console.
        :param dict dataframes_dict: A dictionary that contains Pandas dataframes
                before encoding the features e.g. dataframes_dict={"train": train_dataframe, "test": test_dataframe}
        :param Union[bool, str] _reference: The reference dataframe which is used when applying functions to other dataframes. The default value is the first dataframe inside the dataframes_dict dictionary. Usually it is the train dataframe

        :return:
                - dataframes_dict_encoded - A dictionary that contains Pandas dataframes after encoding the features e.g. dataframes_dict={"train": train_dataframe, "test": test_dataframe}
                - columns_set - A dictionary that contains the features' names sorted in multiple lists based on the type of the data for each given dataset.

        """

        function_id = "1"

        if not _reference:
            _reference = list(dataframes_dict.keys())[0]
        print(f"The reference dataframe is: {_reference}")

        string_columns = self.columns_set[_reference]["categorical_string"] + self.columns_set[_reference][
            "categorical_integer"]

        dataframes_dict_encoded = encode_categorical_features(dataframes_dict, string_columns, print_results)

        print(term.red("*" * 30))

        # After encoding features, A new summary of the data will be presented
        self.columns_set = detect_columns_types_summary(dataframes_dict_encoded, threshold=self.categorical_threshold)

        try:
            self.guidance(self.flow_steps[function_id])
        except Exception as e:
            print(f"It seems that the next step is not defined. Error{e}")

        return dataframes_dict_encoded, self.columns_set

    def scale_data(self, dataframes_dict: dict, ignore_columns: list,
                   _reference: Union[bool, str] = False) -> Tuple[Dict, List]:
        """ Feature scaling

        This function scales features that contains numeric continuous values.

        :param dict dataframes_dict: A dictionary that contains Pandas dataframes
                before scaling features e.g. dataframes_dict={"train": train_dataframe,
                "test": test_dataframe}
        :param list ignore_columns: It contains the columns that should be
                ignored when apply scaling e.g. the id and the target.
        :param Union[bool, str] _reference: The reference dataframe which is used when applying functions to other dataframes. The default value is the first dataframe inside the dataframes_dict dictionary. Usually it is the train dataframe

        :return:
                - dataframes_dict_scaled - A dictionary that contains Pandas dataframes after scaling features e.g. dataframes_dict={"train": train_dataframe, "test": test_dataframe}
                - columns_set - A dictionary that contains the features' names sorted in multiple lists based on the type of the data for each given dataset.
        """

        function_id = "2"

        if not _reference:
            _reference = list(dataframes_dict.keys())[0]

        numerical = [x for x in self.columns_set[_reference]["continuous"] if x not in ignore_columns]

        dataframes_dict_scaled = standard_scale_numeric_features(dataframes_dict, _reference, numerical)

        # Suggesting the next step
        self.columns_set = detect_columns_types_summary(dataframes_dict, threshold=self.categorical_threshold)

        try:
            self.guidance(self.flow_steps[function_id])
        except Exception as e:
            print(f"It seems that the next step is not defined. Error{e}")

        return dataframes_dict_scaled, self.columns_set

    def features_encoding(self, encoding_type: str, dataframe_dict: dict, reference: str,
                          ignore_columns: list, class_number_range: list = None, target_name: str = None,
                          drop_encoded_features: bool = True):
        """ The encoder

        This function encodes the categorical features using different encoding methods
        It assumes that the user encoded the categorical features with string values
        by replacing the those string values by integers

        :param str encoding_type: The type of the encoding method that will be applied. For example: one-hot, target \n
                    For more information please check the following reference:\n
                    https://contrib.scikit-learn.org/categorical-encoding/index.html
        :param dict dataframe_dict: A dictionary that contains Pandas dataframes before the encoding features e.g.
                dataframes_dict={"train": train_dataframe,
                "test": test_dataframe}.
        :param str reference: It is the key ind the dataframes dictionary that points to the dataframe which its
                inputs are taken as a reference to encode the data of other dataframes e.g. "train".
        :param list ignore_columns: It is a list of strings that contains the name of the columns which should be
                ignored when applying the encoding method.
        :param list class_number_range: It is a list of two elements which define the minimum the and maximum number of
                the classes (unique value) that a feature should contain in order to apply the one-hot encoding to
                this feature.
        :param str target_name: The name of the column that contains the labels that should be predicted by the model.
                            If the encoding method doesn't require that target, it can be ignored.
        :param bool drop_encoded_features: If True, the encoded features will be dropped from the dataset

        :return:
                - dataframe_dict_encoded -  A dictionary that contains Pandas dataframes after the encoding features e.g. dataframe_dict_encoded={"train": train_dataframe, "test": test_dataframe}.
                - columns_set - A dictionary that contains the features' names sorted in multiple lists based on the type of the data for each given dataset.
        """

        function_id = f"3-{encoding_type}"

        if class_number_range is None:
            class_number_range = [3, 50]

        categorical_feature = self.columns_set[reference]["categorical_integer"]
        dataframe_dict_encoded = encoding_features(encoding_type,
                                                   dataframe_dict,
                                                   reference,
                                                   categorical_feature,
                                                   ignore_columns,
                                                   class_number_range,
                                                   target_name,
                                                   drop_encoded_features)

        self.columns_set = detect_columns_types_summary(dataframe_dict_encoded, threshold=self.categorical_threshold)

        try:
            self.guidance(self.flow_steps[function_id])
        except Exception as e:
            print(f"It seems that the next step is not defined. Error{e}")

        return dataframe_dict_encoded, self.columns_set

    def training(self, parameters: dict) -> Union[Tuple[List, str], Tuple[List, str, np.array]]:
        """ Ridge linear model

        This function fits the data using the ridge linear model. The function uses the implementation from
        scikit Learn. The user can train a model with specific configuration using the parameter variable.

        :param dict parameters: A dictionary that contains information about the datasets, model type,
                model configurations and training configurations. Check the example below.

        :return:
                - model_index_list - A list of indexes that will be used to point to the trained models which will be saved locally after training
                - save_models_dir - The path where the models will be saved.
                - y_predict - numpy array If the `predict` key is given, the model `my_model` predict the labels of the `parameters["predict"]["test"]` dataset gives back `y_predict`.

        :example:
                >>> parameters = {
                >>>      "data": {
                >>>          "train": {"features": train_dataframe, "target": train_target},
                >>>          "valid": {"features": valid_dataframe, "target": valid_target}, # optional
                >>>          "test": {"features": test_dataframe, "target": test_target}, # optional
                >>>      },
                >>>      "split": {
                >>>          "method": "split",
                >>>          "split_ratios": 0.2,
                >>>      },
                >>>      "model": {"type": "Ridge linear regression",
                >>>                "hyperparameters": {"alpha": 1,
                >>>                                    },
                >>>                },
                >>>      "metrics": ["r2_score", "mean_squared_error"],
                >>>      "predict": { # optional
                >>>          "test": {"features": test_dataframe}
                >>>      }
                >>>  }
        """

        function_id = "4"

        model_index_list, save_models_dir = model_training(parameters)

        if "predict" in parameters.keys():
            model_type = parameters["model"]["type"]
            predict = parameters["predict"]
            y_predict = model_prediction(predict, model_index_list, save_models_dir, model_type)

            try:
                self.guidance(self.flow_steps[function_id])
            except Exception as e:
                print(f"It seems that the next step is not defined. Error{e}")

            return model_index_list, save_models_dir, y_predict
        else:
            return model_index_list, save_models_dir

    @staticmethod
    def comparing_statistics(dataframe_dict: dict):
        """ Datasets statistics visualizer

        This function visualize the statistical properties of the given datasets. It plots Those properties in a
        single graph which help obtain an overview about the distribution of the data in different datasets. It is an
        interactive function. Therefore, it was designed to run in a Jupyter notebook.

        :param dict dataframe_dict: A dictionary that contains Pandas dataframes  e.g. dataframes_dict={"train":
                train_dataframe, "test": test_dataframe}
        """

        compare_statistics(dataframe_dict)

    @staticmethod
    def exploring_data(dataframe_dict: dict, key_i: str):
        """ Datasets explorer

        This functions explore a given dataset by showing information about the most and the least repeated value,
        the number of unique values and the distribution of each feature. It is an interactive function. Therefore,
        it was designed to run in a Jupyter notebook.

        :param dict dataframe_dict: A dictionary that contains Pandas dataframes  e.g. dataframes_dict={"train":
                train_dataframe, "test": test_dataframe}
        :param str key_i: It points to the dataset which the user wants to explore e.g. "train".
        """

        explore_data(dataframe_dict[key_i])

    def flatten_json_data(self, dataframes_dict: dict, _reference: Union[bool, str] = False) -> Tuple[Dict, List]:
        """ JSON data normalizer

        This function normalizes the nested JSON data type inside the pandas dataframes' columns. The name of the new
         columns has the same name of the parent column with a predefined suffix to ensure unique columns' names.

        :param dict dataframes_dict: A dictionary that contains Pandas dataframes with nested JSON data type e.g.
                dataframes_dict={ "train": train_dataframe, "test": test_dataframe}
        :param Union[bool, str] _reference: The reference dataframe which is used when applying functions to other dataframes. The default value is the first dataframe inside the dataframes_dict dictionary. Usually it is the train dataframe


        :return:

            - dataframes_dict - A dictionary that contains Pandas dataframes after flatting the nested JSON data type e.g. dataframes_dict={ "train": train_dataframe, "test": test_dataframe}
            - columns_set - A dictionary that contains the features' names sorted in multiple lists based on the type of the data for each given dataset.

        """

        function_id = "5"

        if not _reference:
            _reference = list(dataframes_dict.keys())[0]

        while len(self.columns_set[_reference]["json"]) > 0:
            dataframes_dict = flat_json(dataframes_dict, self.columns_set[_reference]["json"])
            self.columns_set = detect_columns_types_summary(dataframes_dict, threshold=self.categorical_threshold)

        try:
            self.guidance(self.flow_steps[function_id])
        except Exception as e:
            print(f"It seems that the next step is not defined. Error{e}")

        return dataframes_dict, self.columns_set

    def drop_correlated_columns(self, dataframes_dict: dict, ignore_columns: list, drop_columns: bool = True,
                                print_columns: bool = True, threshold: float = 0.98,
                                _reference: Union[bool, str] = False) -> Tuple[Dict, List]:
        """ Correlation eliminator

        The function drop correlated columns and keep only one of these columns.

        :param dict dataframes_dict: A dictionary that contains Pandas dataframes
                e.g. dataframes_dict={"train": train_dataframe, "test": test_dataframe}
        :param list ignore_columns: It contains the columns that should be ignored e.g. the id and the target.
        :param bool drop_columns: If true, all correlated columns will be dropped but one.
        :param bool print_columns: If True, information about the correlated columns will be printed to the console.
        :param float threshold: A value between 0 and 1. If the correlation between two columns is larger than this.
                value, they are considered highly correlated. If drop_columns is True, one of those columns will be
                dropped. The recommended value of the `threshold` is in [0.7 ... 1].
        :param Union[bool, str] _reference: The reference dataframe which is used when applying functions to other dataframes. The default value is the first dataframe inside the dataframes_dict dictionary. Usually it is the train dataframe


        :return:
                - dataframes_dict - A dictionary that contains Pandas dataframes after dropping correlated columns e.g. dataframes_dict={ "train": train_dataframe, "test": test_dataframe}
                - columns_set - A dictionary that contains the features' names sorted in multiple lists based on the type of the data for each given dataset.
        """

        function_id = "6"

        if not _reference:
            _reference = list(dataframes_dict.keys())[0]
        dataframes_dict[_reference] = drop_corr_columns(dataframes_dict[_reference],
                                                        drop_columns,
                                                        print_columns,
                                                        threshold)
        dataframes_dict = unify_dataframes(dataframes_dict, _reference, ignore_columns)
        self.columns_set = detect_columns_types_summary(dataframes_dict, threshold=self.categorical_threshold)

        try:
            self.guidance(self.flow_steps[function_id])
        except Exception as e:
            print(f"It seems that the next step is not defined. Error{e}")

        return dataframes_dict, self.columns_set

    def drop_columns_constant_values(self, dataframes_dict: dict, ignore_columns: list, drop_columns: bool = True,
                                     print_columns: bool = True,
                                     _reference: Union[bool, str] = False) -> Tuple[Dict, List]:
        """ Constant value features eliminator

        :param dict dataframes_dict: A dictionary that contains Pandas dataframes
                e.g. dataframes_dict={"train": train_dataframe, "test": test_dataframe}
        :param list ignore_columns: It contains the columns that should be ignored e.g. the id and the target.
        :param bool drop_columns: If true, the columns that contain constant values along all the rows will be dropped.
        :param bool print_columns: If true, information about the columns that contain constant values will be printed to the console
        :param Union[bool, str] _reference: The reference dataframe which is used when applying functions to other dataframes. The default value is the first dataframe inside the dataframes_dict dictionary. Usually it is the train dataframe

        :return:
                - dataframes_dict - A dictionary that contains Pandas dataframes after dropping features with constant values e.g. dataframes_dict={ "train": train_dataframe, "test": test_dataframe}
                - columns_set - A dictionary that contains the features' names sorted in multiple lists based on the type of the data for each given dataset.
        """

        function_id = "7"

        if not _reference:
            _reference = list(dataframes_dict.keys())[0]
        dataframes_dict[_reference] = drop_const_columns(dataframes_dict[_reference],
                                                         drop_columns,
                                                         print_columns)

        dataframes_dict = unify_dataframes(dataframes_dict, _reference, ignore_columns)

        self.columns_set = detect_columns_types_summary(dataframes_dict, threshold=self.categorical_threshold)

        try:
            self.guidance(self.flow_steps[function_id])
        except Exception as e:
            print(f"It seems that the next step is not defined. Error{e}")

        return dataframes_dict, self.columns_set

    def update_data_summary(self, dataframes_dict: dict) -> list:
        """Data type updater

        This function update the list of the features in the columns types dictionary. This function should be used in
         case of modifying the features in a  dataset manually. For example, dropping some features or after joining two
         datasets.

        :param dict dataframes_dict: A dictionary that contains Pandas dataframes  e.g. dataframes_dict={"train":
                train_dataframe, "test": test_dataframe}
        :return:
                - columns_set - A dictionary that contains the features' names sorted in multiple lists based on the type of the data for each given dataset.
        """

        function_id = "8"

        self.columns_set = detect_columns_types_summary(dataframes_dict, threshold=self.categorical_threshold)
        try:
            self.guidance(self.flow_steps[function_id])
        except Exception as e:
            print(f"It seems that the next step is not defined. Error{e}")

        return self.columns_set

    def save_post_processed_data(self, dataframe_dict: dict,
                                 path: str,
                                 suffix: str = "",
                                 directory_name: str = "post_processing"):
        """ Data saver

        The function saves data after applying the processing functions.

        :param dict dataframe_dict: A dictionary that contains Pandas dataframes  e.g. dataframes_dict={"train":
                train_dataframe, "test": test_dataframe}
        :param str path: The path to the data.
        :param str suffix: A string that can be added to the filename to get unique names.
        :param str directory_name: The subdirectory where the post-processing data will be saved
        :return:
        """

        function_id = "9"

        path = os.path.join(path, directory_name)
        print(f"Check if the directory {path} exists")

        if not os.path.exists(path):
            os.makedirs(path)

        print(f"{suffix} will be added to the end of the filename")

        filename_list_str = ""

        for key_i, dataframe in dataframe_dict.items():
            filename = key_i + "_" + suffix + ".csv"
            path_i = os.path.join(path, filename)

            if filename_list_str == "":
                filename_list_str = filename
            else:
                filename_list_str = filename_list_str + ", " + filename

            print(f"saving {key_i} dataset to {path_i} file")
            dataframe.to_csv(path_i, index=False)

        try:
            self.guidance(self.flow_steps[function_id])
        except Exception as e:
            print(f"It seems that the next step is not defined. Error{e}")
            FlowInstructions.read_data(specific_directory=path, specific_file_list=f"{filename_list_str}")
        pass



import logging
import os

import ipywidgets as widgets
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import seaborn as sns
from ipywidgets import interact
from matplotlib import colors as m_colors

logger = logging.getLogger(__name__)
formatting = (
    "%(asctime)s: %(levelname)s: File:%(filename)s Function:%(funcName)s Line:%(lineno)d "
    "message:%(message)s"
)
logging.basicConfig(
    filename=os.path.join(os.path.dirname(os.path.dirname(__file__)), "logs/logs.log"),
    level=logging.DEBUG,
    format=formatting,
)


def get_features_set(dataframes_dict: dict):
    features_list = []
    for key_i, dataframe in dataframes_dict.items():
        features_list += list(dataframe.columns)

    features_set = set(features_list)

    return features_set


def histogram(data: pd.Series):
    """ Histogram plotter

    This function plots the histogram of the numeric values of a certain feature when calling the Interactive data
    explorer function in the preprocessing package.

    :param pd.Series data: The values of the given feature/column from the dataset

    """

    _, _, _ = plt.hist(data, 50, facecolor="g", alpha=0.75)
    print(
        "If you applied a normalization function, the x axis will show the normalized values."
    )
    plt.xlabel("Feature's values")
    plt.ylabel("Counts (Frequency)")
    plt.title("Histogram")
    plt.grid(True)
    plt.show()

    pass


def explore_missing_values(dataframes_dict: dict, number_of_features: int):
    """ Missing values explorer

    This function plots the amount of missing values for given amount of features among all datasets.

    :param dict dataframes_dict: A dictionary of pandas dataframes e.g. {"train": train_dataframe,
            "test": test_dataframe}
    :param int number_of_features: The number of the features that should be shown in the plot.

    """

    if number_of_features <= 0:
        logger.error("number of features is invalid")
        raise ValueError("The number of features must be an integer positive and larger than zero")

    maximum_missing_values = 0
    number_of_datasets = len(dataframes_dict.keys())

    logger.debug(f"there are {number_of_datasets} dataframes for exploring the missing values")

    sns.set(style="whitegrid")
    f, ax_set = plt.subplots(number_of_datasets, 1, figsize=(10, 10), sharex=True)

    counter = 0
    g = None

    for key_i, dataframe in dataframes_dict.items():

        missing = dataframe.isnull().sum()
        missing.sort_values(inplace=True, ascending=False)
        missing = missing[:number_of_features]

        logger.debug(f"The number of columns to plot is {number_of_features}")

        g = sns.barplot(x=missing.index, y=missing, palette="deep", ax=ax_set[counter])
        ax_set[counter].axhline(0, color="k", clip_on=False)
        ax_set[counter].set_ylabel(f"{key_i} Count missing values")
        if max(missing) > maximum_missing_values:
            maximum_missing_values = max(missing)

            logger.debug(f"maximum count of missing value is {maximum_missing_values}")

        counter += 1
    # Finalize the plot
    for dataframe_nr in range(len(dataframes_dict.keys())):
        ax_set[dataframe_nr].set(ylim=(0, maximum_missing_values))
    sns.despine(bottom=True)
    plt.tight_layout(h_pad=2)
    g.set_xticklabels(g.get_xticklabels(), rotation=90)

    pass


colors = list(dict(m_colors.BASE_COLORS, **m_colors.CSS4_COLORS).keys())


class CompareStatistics:
    """ Statistic explorer

    The class contains methods for exploring the statistics of a given feature.

    :param dict dataframes_dict: a dictionary of pandas dataframes e.g. {"train": train_dataframe,
            "test": test_dataframe}

    """

    def __init__(self, dataframes_dict: dict):
        """

        :param dataframes_dict: A dictionary of pandas dataframes e.g. {"train": train_dataframe,
            "test": test_dataframe}
        """

        self.dataframes_dict = dataframes_dict
        self.features_set = get_features_set(dataframes_dict)

    def check_feature_valid(self, feature: str) -> bool:
        """ Feature's value validator

        The function validate if it is possible to derive the statistical properties of a given feature.

        :param int feature: The index of the column where the feature is.

        :return:
                True if it is possible to calculate the statistical properties of the given feature. Otherwise false.
        """

        for key_i, dataframe in self.dataframes_dict.items():
            try:
                if (dataframe[feature].dropna().dtype == object) or \
                        (dataframe[feature].dropna().dtype == '<M8[ns]') or \
                        (dataframe[feature].dropna().dtype == '>M8[ns]'):
                    print(
                        f"{feature}: You need to encode the feature's values before plotting statistics"
                    )
                    return False
            except Exception as e:
                print(f"This feature can't be found in all dataframes: {e}")
                return False
        return True

    def compare_statistics_function(self, feature_nr: int):
        """Statistic plotter

        This function plots the statistical values of a certain feature among all given datasets in a single graph.

        :param int feature_nr: The index of the column where the feature is.

        """

        feature = list(self.features_set)[feature_nr]
        if CompareStatistics.check_feature_valid(self, feature):

            fig, ax = plt.subplots()
            counter = 0
            ind = 0

            number_of_dataframes = len(self.dataframes_dict.keys())
            width = 0.7 / number_of_dataframes  # the width of the bars

            for key_i, dataframe in self.dataframes_dict.items():
                try:
                    dataframe_statistic = dataframe.describe().loc[
                                          ['mean', 'std', 'min', '25%', '50%', '75%', 'max'], :]
                    if counter == 0:
                        print(f"{feature}: Comparing the statistical properties")
                        ind = np.arange(len(dataframe_statistic[feature]))  # the x locations for the groups
                    _ = ax.bar(ind - (number_of_dataframes - 1) * width / 2 + counter * width,
                               dataframe_statistic[feature], width,
                               color=colors[counter], label=key_i)
                    counter += 1
                except Exception as e:
                    print(f"The Error is: {e}")

            ax.set_xticklabels(('count', 'mean', 'std', 'min', '25%', '50%', '75%', 'max'))
            ax.legend()


def compare_statistics(dataframes_dict: dict):
    """ Interactive statistic explorer

    This function is designed to be run in a Jupyter notebook. The user can go through the feature interactively using
    a slider.

    :param dict dataframes_dict: A dictionary of pandas dataframes e.g. {"train": train_dataframe,
            "test": test_dataframe}

    """

    compare_statistics_object = CompareStatistics(dataframes_dict)

    maximum_number_features = len(get_features_set(dataframes_dict))

    interact(compare_statistics_object.compare_statistics_function,
             feature_nr=widgets.IntSlider(min=0, max=maximum_number_features - 1, step=1, value=0))



import os
from preprocessing.utils import read_data

path = os.path.join(os.path.dirname(os.path.dirname(__file__)), "data/flow_2")
files_list = ["train.csv", "test.csv"]
test_local = False


# This test has to be applied locally becasuse data is not pushed to Github.
def test_something():
    if test_local:
        dataframe_dict = read_data(path, files_list)
        assert (dataframe_dict["train"].shape[0] == 3000)
        assert (dataframe_dict["test"].shape[0] == 4398)
        assert (dataframe_dict["train"].shape[1] > 1)
        assert (dataframe_dict["test"].shape[1] > 1)
    else:
        pass_test = True
        assert (pass_test == True)



import os
import pandas as pd
import numpy as np
from IPython.display import display


def read_data(path: str, files_list: list, rows_amount: int = 0) -> dict:
    """ CSV file reader

    This function reads CSV files that their names are listed inside the `file_list`

    :param str path: It points to the directory where the data is stored.
    :param list files_list: A list of strings which are the names of the files.
    :param int rows_amount: The number of rows that should be read from the CSV file. If 0, all rows will be read.
    :return:
            dataframes_dictionary: A dictionary that contains Pandas dataframes. The keys are the name of the files
            without the csv extension and the values are the associated dataframes.
    :raise:
            - ValueError - In case of rows_amount has invalid value

    :example:

    >>> path = "./data"
    >>> files_list = ["train.csv", "test.csv"]
    >>> dataframes_dictionary = {"train": train_dataframe, "test": test_dataframe}
    """

    dataframes_dictionary = {}
    for file_i in files_list:
        if rows_amount == 0:
            dataframes_dictionary[file_i.split(".")[0]] = pd.read_csv(os.path.join(path, file_i))
        elif rows_amount > 0:
            try:
                dataframes_dictionary[file_i.split(".")[0]] = pd.read_csv(os.path.join(path, file_i), nrows=rows_amount)
            except Exception as e:
                print(f"Error loading the data using positive rows_amount value. Error: {e}")
        else:
            raise ValueError("rows_amount has invalid value. rows_amount should be 0 or positive number")
    return dataframes_dictionary


def check_if_target_columns_are_imbalanced(dataframes_dict: dict,
                                           possible_problems: dict,
                                           kl_div_threshold: float = 0.05) -> dict:
    """ Detect class imbalance for categorical target variables

    Check if target categorical columns are imbalanced (i.e the different target values do not
    appear with the same frequency).

    :param dict dataframes_dict: dictionary of pandas dataframes e.g. {"train": train_dataframe,
            "test": test_dataframe}
    :param dict possible_problems: dictionary of target candidates associated with the problem type,
            e.g. {"isFraud": "classification", "total amount": "regression"}
    :param float kl_div_threshold:
            threshold above which the Kullback-Leibler Divergence between the ideal and observed distribution is too big.
            The ideal distribution has equal probabilities for all observed values of the target variable.
    :return:
            target_summary dict: summary of value counts for the different categorical target variables.
    """

    def kl_divergence(p, q):
        """ Kullback-Leibler Divergence

        Note that in our case p and q have always non-zero values.
        """
        return np.sum(np.where(p != 0, p * np.log(p / q), 0))

    target_summary = {}
    # at the moment we take into account only classification problems
    try:
        possible_clf_problems = dict((k, v) for k, v in possible_problems.items() if 'classification' == v)
    except Exception as e:
        print("Not possible to check if the target is imbalanced")
        print("Error is:\n", e)
        return target_summary

    # for every target column calculate the counts of distict values
    for target in possible_clf_problems:
        value_counts = [df[target].value_counts().to_frame(name=f"{target} {key_i}")
                        for key_i, df in dataframes_dict.items()
                        if target in df.columns]

        value_counts = pd.concat(value_counts, axis=1).fillna(0)
        value_counts['total count'] = value_counts.sum(axis=1)  # sum value counts from all data frames
        value_counts.index.name = f"{target} values"  # index equal to the distinct target values
        value_counts = value_counts[['total count']]
        value_counts["frequency (%)"] = 100 * value_counts['total count'] / value_counts['total count'].sum()
        value_counts["frequency (%)"] = value_counts["frequency (%)"]

        target_summary[target] = {'content': value_counts}

        # calculate the Kullback-Leibler Divergence for categorical problems
    for k, v in target_summary.items():
        df = v["content"]
        p_ideal = np.ones(df.shape[0]) / df.shape[0]  # ideal case with evenly distributed values
        p_real = df["total count"].values / df["total count"].sum()
        v["kl div"] = kl_divergence(p_ideal, p_real)

    # Check if the values of the target variables are ballanced.
    for k, v in target_summary.items():

        kl_div = v['kl div']

        if kl_div < kl_div_threshold:
            print("\nThe Kullback-Leibler Divergence between probability mass function (pmf) derived from `{0:s}` "
                  "and and an uniformly distributed pmf "
                  " \t = {1:.3f} \nIt is below the threshold {2:.3f}"
                  .format(k, kl_div, kl_div_threshold))
        else:
            print("\nThe Kullback-Leibler Divergence between probability mass function (pmf) derived from `{0:s}` "
                  "and and an uniformly distributed pmf "
                  "\t = {1:.3f} \nIt is above the threshold \t {2:.3f}. \n"
                  "Imballanced target variable!".format(k, kl_div, kl_div_threshold))

        display(v["content"].round(2))

    return target_summary



import logging
import os
from enum import Enum

from IPython.display import display
import pandas as pd

from preprocessing.json_preprocessor import feature_with_json_detector

logger = logging.getLogger(__name__)
formatting = (
    "%(asctime)s: %(levelname)s: File:%(filename)s Function:%(funcName)s Line:%(lineno)d "
    "message:%(message)s"
)
logging.basicConfig(
    filename=os.path.join(os.path.dirname(os.path.dirname(__file__)), "logs/logs.log"),
    level=logging.INFO,
    format=formatting,
)
pd.set_option('display.max_rows', None)


class ColumnDataFormat:
    """ Data type detector

    This class contains methods that help to detect the type of the data in each column of the given dataframe. The
    supported data types for now are: date, categorical feature with string values, numeric features with integer
    values (categorical), numeric with continuous values, and nested JSON data format.

    :param: pd.DataFrame dataframe: A pandas dataframe that contains the dataset e.g. train_dataframe.

    :methods: - `find_date_columns` - Date data type finder
              - `number_or_string` - Numeric-string finder
              - `json_detector`- Valid JSON data finder
              - `categorical_or_numeric` - Numeric continuous-discrete finder
    """

    def __init__(self, dataframe: pd.DataFrame):
        """
        :param pd.DataFrame dataframe:  A pandas dataframe that contains the dataset e.g. train_dataframe
        """

        self.dataframe = dataframe

    def find_date_columns(self) -> list:
        """ Date data type finder

        This method finds date columns automatically.

        :return:
                date_columns: list of columns that contain date format data
        """

        logger.info("Looking for date columns")

        def look_for_date(column_i: pd.Series):
            dates = {date: pd.to_datetime(date) for date in column_i.unique()}
            return column_i.apply(lambda x: dates[x])

        date_columns = []
        possible_date = list(self.dataframe.select_dtypes(include=["datetime"]).columns)
        if possible_date:
            logger.info("Date columns with native date format was found")
            logger.debug(
                f"there are {len(possible_date)} date column with native format (datetime)"
            )

            date_columns = [x for x in possible_date]

            logger.debug(
                f"the columns that contain native date format are {date_columns}"
            )

        for col in self.dataframe.select_dtypes(include=["object"]).columns:
            try:
                self.dataframe[col] = look_for_date(self.dataframe[col])
                date_columns.append(col)
                logger.info(f"column {col} has date data type")
            except ValueError:
                logger.debug(f"{col} has no date data type")
                pass
        return date_columns

    def number_or_string(self, date_columns: list):
        """ Numeric-string finder

        The function extracts which columns in the pandas dataframe contain numeric values and which have string values.
        It returns three lists of strings

        :param list date_columns: contains the name of the columns that have date format data to exclude those
                columns from the search.

        :return:
                - string_columns - A list contains the column names that contain string type data.
                - numeric_columns - contains the list of columns that have numbers.
                - other_columns - contains the name of columns that have unknown type of data if they exist
        """

        string_columns = []
        numeric_columns = []
        other_columns = []

        columns_to_consider = [
            x for x in self.dataframe.columns if x not in date_columns
        ]

        regex_for_numeric = r"[-+]?[0-9]*\.?[0-9]*"
        for column_i in columns_to_consider:

            if self.dataframe[column_i].dropna().dtype == object:
                string_columns.append(column_i)
                continue

            if (
                    self.dataframe[column_i].dropna().astype(str).str.contains(regex_for_numeric, regex=True).all()
            ):
                numeric_columns.append(column_i)
                continue

            other_columns.append(column_i)

        return string_columns, numeric_columns, other_columns

    def json_detector(self, columns_with_strings: list):
        """ Valid JSON data finder

        This method detects if there is valid nested JSON data inside the columns that have string data. It return Two
        lists of strings

        :param list columns_with_strings: List of the columns that contain string data
        :return:
                - string_columns - A list contains the name of the columns that don't have valid JSON nested data.
                - json_columns - A list contain the name of the columns that have valid nested JSON data.
        """

        json_columns = []
        string_columns = []

        for column_i in columns_with_strings:
            if feature_with_json_detector(self.dataframe[column_i]):
                json_columns.append(column_i)
            else:
                string_columns.append(column_i)

        return string_columns, json_columns

    def categorical_or_numeric(self, numbers_column_list: list, threshold: float):
        """ Numeric continuous-discrete finder

        The function extracts the name of the columns that contain numeric discrete values and the columns that have
        numeric continuous values from the columns that contain only number data type. The decision is based on the
        unique values in that column. If the number of the unique values is less than the pre-defined threshold,
        the column type will be considered categorical. It returns two lists of strings.

        :param list numbers_column_list: A list of strings which are the names of the columns that contain number value
                type.
        :param int threshold: It is the minimum number of the unique values that under it the column type will be
                considered categorical.

        :return:
                - categorical_columns - the list of the columns' names of the columns that contain numeric discrete data.
                - numeric_columns - The list of the columns' names that contain numeric continuous data.
        """

        categorical_columns = []
        numeric_columns = []

        for column_i in numbers_column_list:

            if len(self.dataframe[column_i].value_counts()) <= threshold:
                categorical_columns.append(column_i)

            else:
                numeric_columns.append(column_i)

        return categorical_columns, numeric_columns


def detect_column_types(dataframe: pd.DataFrame, threshold: int = 50):
    """ Features' types detector

    This function applies the methods defined in the `ColumnDataFormat` class to detect data format in each column.

    :param pd.DataFrame dataframe: A pandas dataframe that contains the dataset e.g. train_dataframe
    :param int threshold: the minimum number of the unique values that under it the column type will be
            considered categorical. The default value here is 50. This becomes very important when applying one-hot
            encoding.
    :return:
            - number_of_columns - An integer which is the total number of features. It is used for the validation purpose.
            - columns_types_list - A list of lists:
            - string_columns - A list of strings which are the columns that contain categorical data type with string labels  e.g. Yes, No, Maybe.
            - categorical_integer - A list of strings which are the columns that contain categorical data type with numeric labels e.g. 0, 1, 2
            - numeric_columns - A list of strings which are the columns that contain columns contains numeric continuous values e.g. float like 0.1, 0.2 or large number of labels of numeric categorical data (larger than the threshold).
            - date_columns - A list of strings which are the columns that contain columns contain date format data. e.g. 2015-01-05
            - other_columns - A list of strings which are the columns that contain columns that has some other types ( Not implemented yet)
    """

    # create an object
    column_detector = ColumnDataFormat(dataframe)

    # find the date columns first and pass it later to other functions
    date_columns = column_detector.find_date_columns()

    # find columns that have strings or numbers first (other columns will be used later when detecting JSON nested
    # columns)
    string_columns, numeric_columns, other_columns = column_detector.number_or_string(
        date_columns
    )

    # for the numeric columns, find out if they have a large number of unique values or not.
    categorical_integer, numeric_columns = column_detector.categorical_or_numeric(
        numeric_columns, threshold
    )

    # for the string columns, find out if there is valid JSON nested data

    string_columns, json_columns = column_detector.json_detector(string_columns)

    columns_types = {
        "categorical_string": string_columns,
        "categorical_integer": categorical_integer,
        "continuous": numeric_columns,
        "date": date_columns,
        "json": json_columns,
        "other": other_columns
    }

    number_of_columns = sum([len(x) for x in columns_types.values()])

    if number_of_columns != dataframe.shape[1]:
        raise ValueError("Number of columns must be equal to the dataframe's columns")

    return columns_types, number_of_columns


def detect_columns_types_summary(dataframes_dict: dict, threshold: int = 50) -> dict:
    """Data type summarizer

    This function summarize the findings after applying the `detect_column_types` function to each given dataset.

    :param dict dataframes_dict: a dictionary of pandas dataframes e.g. {"train": train_dataframe,
            "test": test_dataframe}
    :param int threshold: The maximum number of categories that a categorical feature should have before considering
            it as continuous numeric feature.
    :return:
            columns_types_dict: A dictionary that contains the lists of the columns filtered based on the type of the
            data that they contain.
    """

    columns_types_dict = {}
    for key_i, dataframe in dataframes_dict.items():
        columns_types, _ = detect_column_types(
            dataframe, threshold=threshold
        )

        columns_types_dict[key_i] = columns_types

    print(f"A summary of the data sets")

    data = dict((k, dict((k_, len(v_)) for k_, v_ in v.items())) for k, v in columns_types_dict.items())

    for k, v in data.items():
        v['total amount'] = sum(v.values())

    summary = pd.DataFrame(data)
    summary.index.name = 'column type'
    display(summary)

    print(
        f"\033[1mNOTE: numeric categorical columns that contains more than {threshold} "
        "classes are considered numeric continuous features.\033[0;0m"
    )
    print(
        "\033[1mNOTE: You can modify the threshold value if you want to consider more or less numeric categorical "
        "features as numeric continuous features.\033[0;0m"
    )

    return columns_types_dict



import logging
import os
import warnings

warnings.filterwarnings('ignore')

import lightgbm as lgb
import pandas as pd
import numpy as np
from sklearn import linear_model
from sklearn.model_selection import KFold, StratifiedKFold
from sklearn.linear_model import LogisticRegression
from IPython.display import display

from training.model_evaluator import evaluate_model, load_all_models
from training.optimizer import get_best_alpha_split, get_best_alpha_kfold
from training.utils import input_parameters_extraction, define_model_directory_name
from training.utils import read_kfold_config, create_model_directory, \
    save_model_locally, split_dataset, xgboost_problem_type
from training.validator import parameters_validator
from training.xgboost_train import training_xgboost_n_split, \
    training_xgboost_kfold, get_num_round, xgboost_data_preparation_for_evaluation
from training.gridsearch_train import train_sklearn_grid_search

logger = logging.getLogger(__name__)
formatting = (
    "%(asctime)s: %(levelname)s: File:%(filename)s Function:%(funcName)s Line:%(lineno)d "
    "message:%(message)s"
)
logging.basicConfig(
    filename=os.path.join(os.path.dirname(os.path.dirname(__file__)), "logs/logs.log"),
    level=logging.INFO,
    format=formatting,
)


def train_with_n_split(test_split_ratios: list, stratify: bool,
                       hyperparameters: dict,
                       train_array: np.array,
                       target: np.array,
                       models_nr: list,
                       model_type: str,
                       required_metrics: list):
    """ n split training

    This function trains a model to fit the data using n split cross-validation e.g train, test or train, valid and test

    :param list test_split_ratios: A list that contains the test split ratio e.g. [0.2] for testing size/training size
            or [0.2, 0.2] for validation size/training size and testing size/(training size - validation size)
    :param bool stratify: If set to True the ratios of the labels is kept the same in the splitted data sets.
    :param dict hyperparameters: A dictionary that contains the hyperparameters which the selected training method
            needs to train the model.
    :param np.array train_array: The values of the features that will be split to two sub-datasets based
            on the split value to multiple datasets.
    :param np.array target: The values of the target that will be split to two sub-datasets based
            on the split value to multiple datasets.
    :param list models_nr: A list of indexes that will be used to point to the trained models which will be saved
            locally after training. In this case there is only one model.
    :param str model_type: The type of model that will be used to fit the data. Currently there are two values:
            Ridge linear regression and lightgbm.
    :param list required_metrics:

    :return:
            - models_nr -  A list of indexes that will be used to point to the trained models which will be saved locally after training. In this case there is only one model.
            - save_models_dir - The name of the directory where the trained models are saved locally.
    """

    problem_to_solve = None
    sub_datasets = split_dataset(train_array, target, test_split_ratios, stratify)

    x_train, y_train = sub_datasets["train"], sub_datasets["target"]

    if model_type == "Ridge linear regression":

        problem_to_solve = "regression"

        logger.info("Ridge linear regression will be used to train the model")
        alpha = hyperparameters["alpha"]

        if alpha == "optimize":
            best_alpha = get_best_alpha_split(x_train, y_train,
                                              sub_datasets["features_0"],
                                              sub_datasets["target_0"])
            linear_regression_model = linear_model.Ridge(alpha=best_alpha)
            print("the optimized alpha value", best_alpha)
        else:
            linear_regression_model = linear_model.Ridge(alpha=alpha)

        model = linear_regression_model.fit(x_train, y_train)

    elif model_type == "lightgbm":

        if hyperparameters["objective"] != "regression":
            problem_to_solve = "classification"
        else:
            problem_to_solve = "regression"

        logger.info("LightGBM regression will be used to train the model")
        lgb_train = lgb.Dataset(x_train, label=y_train.reshape(len(y_train)))

        lgb_valid = lgb.Dataset(sub_datasets["features_0"],
                                label=sub_datasets["target_0"].reshape(len(sub_datasets["target_0"])))

        # start training the model
        model = lgb.train(hyperparameters, train_set=lgb_train, valid_sets=[lgb_valid], verbose_eval=0)

    elif model_type == "Logistic regression":
        problem_to_solve = "classification"
        logger.info("Logistic regression will be used to train the model")
        if len(set(y_train)) > 2:
            logistic_regression_model = LogisticRegression(multi_class='multinomial')
        else:
            logistic_regression_model = LogisticRegression()

        model = logistic_regression_model.fit(x_train, y_train)

    elif model_type == "xgboost":
        num_round = get_num_round(hyperparameters)
        model, problem_to_solve, validation_list = training_xgboost_n_split(sub_datasets, hyperparameters, num_round)

    elif model_type.split(".")[0] == "sklearn":
        model = train_sklearn_grid_search(x_train, y_train, model_type, hyperparameters)
        problem_to_solve = hyperparameters["objective"]
    else:
        logger.error("Model type is not recognized")
        raise ValueError("Model type is not recognized")

    # Evaluate the model on all sub-datasets
    metrics_summary_all = dict()
    metrics_summary_all["model 0"] = evaluate_model(
        model,
        xs=[x_train] + [sub_datasets[f"features_{i}"] for i in range(len(test_split_ratios))],
        ys=[y_train] + [sub_datasets[f"target_{i}"] for i in range(len(test_split_ratios))],
        labels=["(train.train)"] + [f"(train.validation_{i})" for i in range(len(test_split_ratios))],
        metrics=required_metrics)

    metrics_summary_all = pd.DataFrame(metrics_summary_all)
    display(metrics_summary_all)

    # Define the directory name of the models
    split = "n_split"
    save_models_dir = define_model_directory_name(model_type, hyperparameters, split, problem_to_solve)
    # Get a unique name when saving the model
    for n_split in test_split_ratios:
        save_models_dir += f'_{str(n_split).replace(".", "_")}'
    logger.debug(f"the name of the directory where the model will be saved is: {save_models_dir}")
    create_model_directory(save_models_dir)

    path = os.path.join(save_models_dir, f"{model_type}_{0}.pkl")
    save_model_locally(path, model)

    # we have only one model in this case
    models_nr.append("0")

    logger.info("Training the model using split cross-validation is finished")
    return models_nr, save_models_dir


def train_with_kfold_cross_validation(split: dict, stratify: bool,
                                      hyperparameters: dict,
                                      train_array: np.array,
                                      target: np.array,
                                      models_nr: list,
                                      model_type,
                                      required_metrics: list):
    """K-Fold cross-validation training

    This function trains a model to fit the data using K-Fold cross-validation.

    :param dict split: A dictionary that contains information about the K-Fold variables
    :param bool stratify: If set to True the ratios of the labels is kept the same in the splitted data sets.
    :param dict hyperparameters: A dictionary that contains the hyperparameters which the selected training method
            needs to train the model.
    :param np.array train_array: The values of the target that will be split into K-Folds and used to train the model to
            predict the target
    :param np.array target: The values of the target that will be split into K-Folds and used to train the model.
    :param list models_nr: A list of indexes that will be used to point to the trained models which will be saved
            locally after training. In this case there are n_fold models.
    :param str model_type: The type of model that will be used to fit the data.
    :param list required_metrics:

    :return:
            - models_nr - A list of indexes that will be used to point to the trained models which will be saved locally after training. In this case there are n_fold models.
            - save_models_dir - The name of the directory where the trained models are saved locally.
    """
    problem_to_solve = None

    # Ensure that the user provided the required variable.
    n_fold, shuffle, random_state = read_kfold_config(split)

    if stratify:
        kfold = StratifiedKFold(n_splits=n_fold, shuffle=shuffle, random_state=random_state)
    else:
        kfold = KFold(n_splits=n_fold, shuffle=shuffle, random_state=random_state)

    # Avoiding warning of local variable might be referenced before assignment
    linear_regression_model = None
    logistic_regression_model = None

    if model_type == "Ridge linear regression":
        problem_to_solve = "regression"
        alpha = hyperparameters["alpha"]
        if alpha == "optimize":
            alpha = get_best_alpha_kfold(kfold, train_array, target)
        linear_regression_model = linear_model.Ridge(alpha=alpha)
    elif model_type == "lightgbm":
        if hyperparameters["objective"] != "regression":
            problem_to_solve = "classification"
        else:
            problem_to_solve = "regression"
    elif model_type == "Logistic regression":
        problem_to_solve = "classification"
        if len(set(target)) > 2:
            logistic_regression_model = LogisticRegression(multi_class='multinomial')
        else:
            logistic_regression_model = LogisticRegression()
    elif model_type == "xgboost":
        # Find out what is the type of the problem that should be solved based on the defined objective
        problem_to_solve = xgboost_problem_type(hyperparameters)
    elif model_type.split(".")[0] == "sklearn":
        problem_to_solve = hyperparameters["objective"]

    fold_nr = 0  # counter for identifying models
    metrics_summary_all = {}
    # Define the directory name of the models
    split = "kfold"
    save_models_dir = define_model_directory_name(model_type, hyperparameters, split, problem_to_solve)
    save_models_dir += f"_folds_n_{n_fold}"
    create_model_directory(save_models_dir)

    for train, test in kfold.split(train_array, target):
        fold_nr += 1
        print("fold_nr.", fold_nr)

        if model_type == "Ridge linear regression":
            logger.info("LightGBM regression will be used to train the model")
            kfold_model = linear_regression_model.fit(train_array[train], target[train])

        elif model_type == "lightgbm":
            logger.info("LightGBM will be used to train the model")
            lgb_train = lgb.Dataset(train_array[train], label=target[train].reshape(len(target[train])))
            lgb_valid = lgb.Dataset(train_array[test], label=target[test].reshape(len(target[test])))

            # start training the model
            kfold_model = lgb.train(hyperparameters, train_set=lgb_train, valid_sets=[lgb_valid], verbose_eval=0)

        elif model_type == "Logistic regression":
            kfold_model = logistic_regression_model.fit(train_array[train], target[train])

        elif model_type == "xgboost":
            num_round = get_num_round(hyperparameters)
            kfold_model, validation_list = training_xgboost_kfold(train_array,
                                                                  target,
                                                                  train,
                                                                  test,
                                                                  hyperparameters,
                                                                  num_round)

        else:
            logger.error("Model type is not recognized")
            raise ValueError("Model type is not recognized")

        path = os.path.join(save_models_dir, f"{model_type}_{fold_nr}.pkl")
        save_model_locally(path, kfold_model)

        models_nr.append(fold_nr)

        if model_type == "xgboost":
            metrics_summary = evaluate_model(kfold_model,
                                             xs=[validation_list[0][0], validation_list[1][0]],
                                             ys=[target[train], target[test]],
                                             labels=["(train.train)", "(train.validation)"],
                                             metrics=required_metrics)
        else:
            metrics_summary = evaluate_model(kfold_model,
                                             xs=[train_array[train], train_array[test]],
                                             ys=[target[train], target[test]],
                                             labels=["(train.train)", "(train.validation)"],
                                             metrics=required_metrics)

        metrics_summary_all[f"fold_{fold_nr}"] = metrics_summary
        print(metrics_summary)

    metrics_summary_all = pd.DataFrame(metrics_summary_all)
    metrics_summary_all['mean'] = metrics_summary_all.mean(axis=1)
    display(metrics_summary_all)

    logger.info("Training the model using KFold cross-validation is finished")
    return models_nr, save_models_dir


def model_training(parameters: dict):
    """ Model training

    This function trains a model to fit the data using the Scikit Learn of Ridge linear model implementation

    :param dict parameters: A dictionary that contains information about the datasets, model type,
                model configurations and training configurations. Check the example below.
    :return:
            - models_nr - A list of indexes that will be used to point to the trained models which will be saved locally after training.
            - save_models_dir - The name of the directory where the trained models are saved locally.
    :example:
            One split: train and test
                    >>> parameters = {
                    >>>      "data": {
                    >>>          "train": {"features": train_dataframe, "target": train_target},
                    >>>          "valid": {"features": valid_dataframe, "target": valid_target}, # optional
                    >>>          "test": {"features": test_dataframe, "target": test_target}, # optional
                    >>>      },
                    >>>      "split": {
                    >>>          "method": "split",
                    >>>          "split_ratios": 0.2,
                    >>>      },
                    >>>      "model": {"type": "Ridge linear regression",
                    >>>                "hyperparameters": {"alpha": 1,
                    >>>                                    },
                    >>>                },
                    >>>      "metrics": ["r2_score", "mean_squared_error"],
                    >>>      "predict": { # optional
                    >>>          "test": {"features": test_dataframe}
                    >>>      }
                    >>>  }
            Two splits: train, valid and test
                    >>> parameters = {
                    >>>      "data": {
                    >>>          "train": {"features": train_dataframe, "target": train_target},
                    >>>          "valid": {"features": valid_dataframe, "target": valid_target}, # optional
                    >>>          "test": {"features": test_dataframe, "target": test_target}, # optional
                    >>>      },
                    >>>      "split": {
                    >>>          "method": "split",
                    >>>          "split_ratios": (0.2, 0.2), # or [0.2, 0.2]
                    >>>      },
                    >>>      "model": {"type": "Ridge linear regression",
                    >>>                "hyperparameters": {"alpha": 1,
                    >>>                                    },
                    >>>                },
                    >>>      "metrics": ["r2_score", "mean_squared_error"],
                    >>>      "predict": { # optional
                    >>>          "test": {"features": test_dataframe}
                    >>>      }
                    >>>  }
            KFold cross-validation:
                    >>> parameters = {
                    >>>      "data": {
                    >>>          "train": {"features": train_dataframe, "target": train_target},
                    >>>          "valid": {"features": valid_dataframe, "target": valid_target}, # optional
                    >>>          "test": {"features": test_dataframe, "target": test_target}, # optional
                    >>>      },
                    >>>      "split": {
                    >>>          "method": "kfold",
                    >>>          "fold_nr": 5,
                    >>>      },
                    >>>      "model": {"type": "Ridge linear regression",
                    >>>                "hyperparameters": {"alpha": 1,
                    >>>                                    },
                    >>>                },
                    >>>      "metrics": ["r2_score", "mean_squared_error"],
                    >>>      "predict": { # optional
                    >>>          "test": {"features": test_dataframe}
                    >>>      }
                    >>>  }
            KFold cross-validation with alpha optimization:
                    >>> parameters = {
                    >>>      "data": {
                    >>>          "train": {"features": train_dataframe, "target": train_target},
                    >>>          "valid": {"features": valid_dataframe, "target": valid_target}, # optional
                    >>>          "test": {"features": test_dataframe, "target": test_target}, # optional
                    >>>      },
                    >>>      "split": {
                    >>>          "method": "kfold",
                    >>>          "fold_nr": 5,
                    >>>      },
                    >>>      "model": {"type": "Ridge linear regression",
                    >>>                "hyperparameters": {"alpha": "optimize",
                    >>>                                    },
                    >>>                },
                    >>>      "metrics": ["r2_score", "mean_squared_error"],
                    >>>      "predict": { # optional
                    >>>          "test": {"features": test_dataframe}
                    >>>      }
                    >>>  }

    """

    parameters_validator(parameters)
    data, split, train_array, target, model_type, hyperparameters, predict = input_parameters_extraction(parameters)
    stratify = parameters["split"].get("stratify", False)

    models_nr = []
    save_models_dir = None

    if split["method"] == "split":
        logger.info("Start training using Split cross-validation")
        if type(split["split_ratios"]) == float:
            test_split_ratio = [split["split_ratios"]]
        else:
            test_split_ratio = split["split_ratios"]

        models_nr, save_models_dir = train_with_n_split(test_split_ratio,
                                                        stratify,
                                                        hyperparameters,
                                                        train_array,
                                                        target,
                                                        models_nr,
                                                        model_type,
                                                        parameters["metrics"])

    elif split["method"] == "kfold":
        logger.info("Start training using KFold cross-validation")
        models_nr, save_models_dir = train_with_kfold_cross_validation(split,
                                                                       stratify,
                                                                       hyperparameters,
                                                                       train_array,
                                                                       target,
                                                                       models_nr,
                                                                       model_type,
                                                                       parameters["metrics"])

    # Evaluate models
    if model_type == "xgboost":
        data = xgboost_data_preparation_for_evaluation(data)

    metrics_summary_all = {}
    for model_i in models_nr:
        model = load_all_models(save_models_dir, model_type, model_i)
        xs = [v["features"] for k, v in data.items()]
        ys = [v["target"] for k, v in data.items()]
        labels = [f"({k})" for k, v in data.items()]

        metrics_summary = evaluate_model(model, xs, ys, labels, parameters["metrics"])

        metrics_summary_all[f"model {model_i}"] = metrics_summary

    metrics_summary_all = pd.DataFrame(metrics_summary_all)
    if metrics_summary_all.shape[1] > 1:
        # calculate mean only if we have more than one model
        metrics_summary_all['mean'] = metrics_summary_all.mean(axis=1)
    display(metrics_summary_all)

    logger.info("Training process is finished")
    return models_nr, save_models_dir



import unittest
import numpy as np
import pandas as pd
from training.xgboost_train import xgboost_data_preparation, xgboost_regression_train


# create dataframe for testing the preprocessing functions:
def mock_data(number_of_samples):
    integer_array = np.random.randint(2, size=(number_of_samples, 2))
    for categories_numbers in range(5, 50, 10):
        integer_array = np.append(
            integer_array, np.random.randint(categories_numbers, size=(number_of_samples, 2)), axis=1
        )
    integer_columns = [f"int_col_{x}" for x in range(integer_array.shape[1])]

    continuous_array = np.random.randn(number_of_samples, 10)
    continuous_columns = [f"cont_col_{x}" for x in range(continuous_array.shape[1])]

    integer_dataframe = pd.DataFrame(integer_array, columns=integer_columns)
    continuous_dataframe = pd.DataFrame(continuous_array, columns=continuous_columns)

    dataframe = pd.concat([integer_dataframe, continuous_dataframe], axis=1)
    target = (np.sum(continuous_array, axis=1) - 1) / (1 + np.sum(integer_array, axis=1))
    return dataframe, target


train_dataframe, train_target = mock_data(1000)
valid_dataframe, valid_target = mock_data(100)
test_dataframe, test_target = mock_data(100)

parameters_linear = {
    "data": {
        "train": {"features": train_dataframe, "target": train_target},
        "valid": {"features": valid_dataframe, "target": valid_target},
        "test": {"features": test_dataframe, "target": test_target},
    },
    "split": {
        "method": "split",  # "method":"kfold"
        "split_ratios": 0.2,  # foldnr:5 , "split_ratios": 0.8 # "split_ratios":(0.7,0.2)
    },
    "model": {"type": "Ridge linear regression",
              "hyperparameters": {"alpha": 1,  # alpha:optimize
                                  },
              },
    "metrics": ["r2_score", "mean_squared_error"],
    "predict": {
        "test": {"features": test_dataframe}
    }
}

validation_list = []
for data_i in parameters_linear["data"].keys():
    dataframe = parameters_linear["data"][data_i]["features"]
    target = parameters_linear["data"][data_i]["target"]

    validation_list = xgboost_data_preparation(validation_list, dataframe, target, data_i)

assert isinstance(validation_list, list)
assert len(validation_list) == 3

hyperparameters = {'max_depth': 5, 'eta': 1, 'eval_metric': "rmse"}
bst = xgboost_regression_train(validation_list, hyperparameters, num_round=10)
print(type(bst))
assert isinstance(bst, object)



import unittest
import numpy as np
import pandas as pd
from datetime import datetime, timedelta

from preprocessing.data_science_help_functions import detect_id, detect_target, detect_problem_type

# create dataframe for testing the preprocessing functions:
integer_array = np.random.randint(2, size=(100, 2))
for catagories_numbers in range(5, 50, 10):
    integer_array = np.append(
        integer_array, np.random.randint(catagories_numbers, size=(100, 2)), axis=1
    )
integer_columns = [f"int_col_{x}" for x in range(integer_array.shape[1])]

continuous_array = np.random.randn(100, 10)
continuous_columns = [f"cont_col_{x}" for x in range(continuous_array.shape[1])]

string_array = [
    "pizza",
    "ball",
    "star3",
    "car",
    "01515",
    "cat75",
    "moon",
    "lol",
    "dddd",
    "wtf",
    "xXx",
    "82008",
    "mailbox",
]
string_arrays = np.random.choice(string_array, (100, 10))
string_columns = [f"string_col_{x}" for x in range(string_arrays.shape[1])]

integer_dataframe = pd.DataFrame(integer_array, columns=integer_columns)
continuous_dataframe = pd.DataFrame(continuous_array, columns=continuous_columns)
string_dataframe = pd.DataFrame(string_arrays, columns=string_columns)

date_array = np.array([datetime.now() - timedelta(days=x) for x in range(100)]).reshape(
    100, 1
)
date_dataframe = pd.DataFrame(date_array, columns=["date_col"])
# print(date_dataframe.dtypes)
train_dataframe = pd.concat(
    [date_dataframe, integer_dataframe, continuous_dataframe, string_dataframe], axis=1
)

dataframe_dict = {
    "train": train_dataframe,
    "valid": train_dataframe,
    "test": train_dataframe,
}

id_set = detect_id(dataframe_dict)
target_set = detect_target(dataframe_dict)
problem_type_dict = detect_problem_type(dataframe_dict, target_set)


class TooManyOptions(unittest.TestCase):
    def test_train_dataset(self):
        self.assertEqual(id_set, "Too many options for ids. Not possible to detect id")
        self.assertEqual(target_set, "Too many options for target. Not possible to detect target")
        self.assertEqual(problem_type_dict, "No problem type to detect")


dataframe_dict = {
    "train": train_dataframe,
}

id_set_2 = detect_id(dataframe_dict)
print(id_set_2)
target_set_2 = detect_target(dataframe_dict)
print(target_set_2)
problem_type_dict_2 = detect_problem_type(dataframe_dict, target_set_2)
print(problem_type_dict_2)


class OneDataFrame(unittest.TestCase):
    def test_train_dataset(self):
        self.assertEqual(id_set_2, "Too many options for ids. Not possible to detect id")
        self.assertEqual(target_set_2, "Too many options for target. Not possible to detect target")
        self.assertEqual(problem_type_dict_2, "No problem type to detect")



if __name__ == "__main__":
    unittest.main()



import logging
import os
from typing import Tuple, List, Dict, Set, Union
from collections import Counter

logger = logging.getLogger(__name__)
formatting = (
    "%(asctime)s: %(levelname)s: File:%(filename)s Function:%(funcName)s Line:%(lineno)d "
    "message:%(message)s"
)
logging.basicConfig(
    filename=os.path.join(os.path.dirname(os.path.dirname(__file__)), "logs/logs.log"),
    level=logging.INFO,
    format=formatting,
)


def detect_id(dataframes_dictionary: dict) -> Union[set, str]:
    """ ID candidates detector

    the following assumptions are considered:

    | 1. Id exists in all the loaded datasets;
    | 2. All the id' values are unique in all datasets;
    | 3. There are no missing values.

    If the function finds more than one feature that satisfies the assumptions, it suggests a list of candidates
    and the user has to decide which one is the id

    :param dict dataframes_dictionary: It is a dictionary that contains pandas dataframes e.g. dataframes_dictionary ={
    'train': train_dataframe, 'test': test_dataframe}

    :return A set of strings values as id candidates.

    :rtype: set
    """

    id_candidates = []

    for key_i, dataframe in dataframes_dictionary.items():
        for column_i in dataframe.columns:
            if len(dataframe[column_i].value_counts()) == dataframe[column_i].count() and \
                    len(dataframe[column_i].value_counts()) == dataframe.shape[0]:
                logger.info(f"{column_i} found to be a candidate as an ID")

                id_candidates.append(column_i)

    id_candidates = set(id_candidates)

    # the number 5 in the if statement should be defined as variable later that can be added by the user
    if len(id_candidates) > 5:
        logger.info("Too many options for ids.")
        return "Too many options for ids. Not possible to detect id"
    elif len(id_candidates) == 0:
        logger.info("No ids candidates were found")
        return "No ids candidates were found"

    logger.info("Detecting the id is finished")

    return set(id_candidates)


def detect_target(dataframes_dictionary: dict) -> Union[list, str]:
    """ Target candidates detector

    The following assumptions are considered:

    | 1. Target is missing in at least one dataset-which is the test dataset usually;
    | 2. The target has no missing values.

    If the function finds more than one feature that satisfies the assumptions, it suggests a list of candidates
    and the user has to decide which one is the target

    :param dict dataframes_dictionary: It is a dictionary that contains pandas dataframes e.g. dataframes_dictionary ={
    'train': train_dataframe, 'test': test_dataframe}

    :return
        target_candidates_2: A list of candidates as strings that satisfies the two mentioned assumptions above

    :rtype: list

    """

    target_candidates_1 = []
    target_candidates_2 = []
    columns = []

    for key_i, dataframe in dataframes_dictionary.items():
        # Getting all columns names of all datasets in one place
        columns = columns + list(dataframe.columns)

    # check the first assumptions: calculating the occurrence
    try:
        occurrence = Counter(columns)
        minimum_occurrence = min(occurrence.values())
        target_candidates_1 = [x for x, y in occurrence.items() if y == minimum_occurrence]
        logger.info("Checking the first assumption (occurrence) for finding the target is done!")

    except Exception as e:
        logger.error(f"Error: {e}")

    # check the second assumption: No missing values
    for column_i in target_candidates_1:
        for key_i, dataframe in dataframes_dictionary.items():
            if column_i in dataframe.columns:
                if len(dataframe[column_i]) - dataframe[column_i].count() == 0:
                    target_candidates_2.append(column_i)

    logger.info("Checking the second assumption (missing values) for finding the target is done!")

    if len(target_candidates_2) == 0:
        logger.info("No target was detected")
        return "No target was detected"
    elif len(target_candidates_2) > 5:
        logger.info("Too many options for target")
        return "Too many options for target. Not possible to detect target"

    logger.info("Detecting the target is finished")

    return target_candidates_2


def detect_problem_type(dataframes_dictionary: dict,
                        target_candidates: list,
                        threshold: float = 0.1) -> Union[dict, str]:
    """ Problem type detector

    This function tells what type of problem that should be solved: classification or regression

    :param dict dataframes_dictionary: A dictionary that contains pandas dataframes e.g. dataframes_dictionary ={
                                        'train': train_dataframe, 'test': test_dataframe}
    :param list target_candidates: It is list of string of the possible target candidates
    :param float threshold: A value larger than 0 and less than 1. It defines when the problem is considered as a
    regression  problem or classification problem

    :return:
            problem_type: dictionary of target candidates associated with the problem type

    :rtype: dict
    """

    problem_type = {}

    if isinstance(target_candidates, list) and len(target_candidates) > 0:
        for column_i in target_candidates:
            for key_i, dataframe in dataframes_dictionary.items():
                if column_i in dataframe.columns:
                    if len(dataframe[column_i].value_counts()) / dataframe.shape[0] < threshold:
                        problem_type[column_i] = "classification"
                        logger.debug(f"For the target {column_i}: classification")
                    else:
                        problem_type[column_i] = "regression"
                        logger.debug(f"For the target {column_i}: regression")

    else:
        logger.info("No valid target candidates")
        return "No problem type to detect"

    logger.info("Detecting the problem type is finished")

    return problem_type


def detect_id_target_problem(dataframes_dict: dict, threshold: float = 0.1) -> Tuple[Set, List, Dict]:
    """ ID Target Problem type detector

    This function tries to find which column is the ID and which one is the target and what type of the problem
    to be solved. It uses `detect_id`, `detect_target` and `detect_problem_type` functions.

    :param dict dataframes_dict: A dictionary that contains pandas dataframes e.g. dataframes_dictionary ={
                'train': train_dataframe, 'test': test_dataframe}
    :param float threshold: A value larger than 0 and less than 1. It defines when the problem is a regression
            problem or classification problem

    :return:
            | possible_ids: A list of candidates as strings.
            | possible_target: list of candidates as strings.
            | possible_problems: dictionary of target candidates associated with the problem type.
    """

    possible_ids = detect_id(dataframes_dict)
    possible_target = detect_target(dataframes_dict)
    possible_problems = detect_problem_type(dataframes_dict, possible_target, threshold=threshold)

    logger.info("Running all the process to detect id, target and the problem type is finished")

    # Showing the info to the user
    print(f"The possible ids are:\n {possible_ids}")
    print(f"The possible possible_target are:\n {possible_target}")
    print(f"The type of the problem that should be solved:\n {possible_problems}")
    return possible_ids, possible_target, possible_problems



import logging
import os

import pandas as pd

logger = logging.getLogger(__name__)
formatting = (
    "%(asctime)s: %(levelname)s: File:%(filename)s Function:%(funcName)s Line:%(lineno)d "
    "message:%(message)s"
)
logging.basicConfig(
    filename=os.path.join(os.path.dirname(os.path.dirname(__file__)), "logs/logs.log"),
    level=logging.INFO,
    format=formatting,
)


def drop_corr_columns(dataframe: pd.DataFrame, drop_columns: bool = True,
                      print_columns: bool = True, threshold: float = 0.98) -> pd.DataFrame:
    """ Correlated columns eliminator

    The function drop correlated columns and keep only one of these columns. Usually removing high correlated columns
    gives improvement in model's quality. The task of this function is first to print list of the most correlated
    columns and then remove them by threshold. For more information, please refer to pandas.DataFrame.corr description:
    https://pandas.pydata.org/pandas-docs/stable/reference/api/pandas.DataFrame.corr.html

    :param pd.DataFrame dataframe: Pandas dataframe which contains the dataset e.g. train_dataframe.
    :param bool drop_columns: If true, all correlated columns will be dropped but one.
    :param bool print_columns: If True, information about the correlated columns will be printed to the console.
    :param float threshold: A value between 0 and 1. If the correlation between two columns is larger than this.
                            value, they are considered highly correlated. If drop_columns is True, one of those columns will be dropped. The
                            recommended value of the `threshold` is in [0.7 ... 1].
    
    :return:
            dataframe: A pandas dataframe which contains the dataset after dropping the correlated columns if `drop_columns = True`. Otherwise, the same input dataframe will be returned.
    
    :example:
    
    For checking correlated columns:
    
    >>> dataframe = drop_corr_columns(dataframe, drop_columns=False, print_columns=True, threshold=0.85)
    """

    # 1. calculation
    logger.info("Calculating the correlation matrix")

    correlation_coefficients = dataframe.corr()

    # 2. report
    corr_fields_list = []
    print(f"Columns with correlations more than {str(threshold)} :")
    for i in correlation_coefficients:
        for j in correlation_coefficients.index[correlation_coefficients[i] >= threshold]:
            if i != j and j not in corr_fields_list:
                corr_fields_list.append(j)

                logger.info("Report information if required")

                if print_columns:
                    logger.debug(f"print_columns = {print_columns}: Information should be reported")
                    print(
                        f"{i}-->{j}: r^2={correlation_coefficients[i][correlation_coefficients.index == j].values[0]}"
                    )

    # 3. dropping
    logger.info("Dropping high correlated columns if required")

    if drop_columns:
        logger.debug(f"drop_columns = {drop_columns}: Columns should be dropped")

        print(f"{dataframe.shape[1]} columns total")
        dataframe = dataframe.drop(corr_fields_list, 1)
        print(f"{dataframe.shape[1]} columns left")

    return dataframe


def drop_const_columns(dataframe: pd.DataFrame, drop_columns: bool = True, print_columns: bool = True) -> pd.DataFrame:
    """ Constant value columns eliminator

    This function drops columns that contain constant values. Usually removing constant columns gives improvement in
    model's quality. The task of this function is first to print list of constant columns and then drop them.

    :param pd.DataFrame dataframe: A pandas dataframe that contain the dataset e.g. train_dataframe
    :param bool drop_columns: If true, the columns that contain constant values along all the rows will be dropped.
    :param bool print_columns: If true, information about the columns that contain constant values will be printed to the console

    :return:
            dataframe: A pandas dataframe that contains the dataset after dropping the the columns that contain
            constant values if `drop_columns = True`

    :example:

    For checking the columns which have constant value:

    >>> dataframe = drop_const_columns(dataframe, drop_columns=False, print_columns=True)
    """

    # 1. report
    single_value_cols = []
    for col in dataframe.columns:
        unique_count = dataframe[col].nunique()
        if unique_count < 2:
            single_value_cols.append(col)

            logger.info("Calculating the correlation matrix")

            if print_columns:
                logger.debug(f"print_columns = {print_columns}: Information should be reported")

                print(col, unique_count)

    print(f"Constant columns count: {len(single_value_cols)}")

    # 2. dropping
    logger.info("Dropping high correlated columns if required")

    if drop_columns:
        logger.debug(f"drop_columns = {drop_columns}: Columns should be dropped")

        print(f"{dataframe.shape[1]} columns total")
        dataframe = dataframe.drop(single_value_cols, 1)
        print(f"{dataframe.shape[1]} columns left")

    return dataframe



import logging
import os

import ipywidgets as widgets
import numpy as np
import pandas as pd
from ipywidgets import interact

from visualization.visualization import histogram

logger = logging.getLogger(__name__)
formatting = (
    "%(asctime)s: %(levelname)s: File:%(filename)s Function:%(funcName)s Line:%(lineno)d "
    "message:%(message)s"
)
logging.basicConfig(
    filename=os.path.join(os.path.dirname(os.path.dirname(__file__)), "logs/logs.log"),
    level=logging.INFO,
    format=formatting,
)


def print_repeated_values(series_data: pd.Series):
    """ Repeated values displayer

     This function prints out into the console the results of value_counts. It shows count_values.head() or tail()

    :param pd.Series series_data: It the values of one of the features that are in the given dataset.

    """

    try:
        max_space = max([len(str(x)) for x in series_data.index])

        logging.debug(f"The maximum value's length in the given series is {max_space}")

        for value_nr in range(len(series_data)):
            index_i = series_data.index[value_nr]
            print(
                f"the value {index_i} {' ' * (3 + max_space - len(str(index_i)))} is found "
                f"{'' * (1 + max_space - len(str(series_data[index_i])))}"
                f" {series_data[index_i]} times"
            )
    except Exception as e:
        logger.error(f"Error: {e}")


class ExploreData:
    """ Data explorer

    This class have the `data explore` method which can be used to explore the data in each column in the dataset.

    :param pd.DataFrame dataframe: A pandas dataframe that contains the dataset e.g. train_dataframe

    """

    def __init__(self, dataframe: pd.DataFrame):
        """

        :param pd.DataFrame dataframe: A pandas dataframe that contains the dataset e.g. train_dataframe
        """

        self.dataframe = dataframe

    def data_explore(self, column_i: str):
        """ Feature explorer

        This method displays a summary about the given feature including missing values, most and least repeated
        values. Besides that, it shows the histogram of the numeric data.

        :param str column_i: The name of the feature that the user is interested in exploring.

        """

        column_name = self.dataframe.columns[column_i]
        logger.info(f"Exploring the values of the feature {column_name}")

        print(f"\033[1mExploring column: {column_name}\033[0;0m")

        if self.dataframe[column_name].dtype == object:
            logger.debug(f"The type of the data of the feature {column_name} is object")

            # To take the missing value into account, I replace the missing values with "missing values"
            count_values = self.dataframe[column_name].fillna("missing values").value_counts(dropna=False)

            print("If there are missing values, they will be replaced by the term: missing values")

        elif self.dataframe[column_name].dtype == float:
            logger.debug(f"The type of the data of the feature {column_name} is float")

            # To take the missing value into account, I replace the missing values with the mean values
            count_values = self.dataframe[column_name].fillna(np.nanmean(self.dataframe[column_name])).value_counts(
                dropna=False)

            print(f"If there are missing values," +
                  f" they will be replaced by the mean: {np.nanmean(self.dataframe[column_name])}")
        else:
            logger.debug(f"The type of the data of the feature {column_name} is something else than object or "
                         "float-most probably categorical with integer values")

            # To take the missing value into account, I replace the missing values with the value -9999
            count_values = self.dataframe[column_name].fillna(-9999).value_counts(dropna=False)
            print("If there are missing values, they will be replaced by the term: -9999")

        # showing information to the user
        print("The 5 most repeated values:")
        print_repeated_values(count_values.head())
        print("The 5 least repeated values")
        print_repeated_values(count_values.tail())
        print("The number of unique values is -->", len(count_values))

        # In order to show the histogram, the values should be numeric
        if self.dataframe[column_name].dropna().dtype == object:
            logger.debug(f"It is not possible to show the histogram of the feature {column_name}")
            print(
                "You need to encode the feature's values before showing the histogram"
            )
        else:
            logger.debug(f"The feature {column_name} has bool type values")

            # For showing the histogram of the bool values, the True, False values should be replaced by 1, 0 values.
            if self.dataframe[column_name].dropna().dtype == bool:
                tmp_dataframe = self.dataframe[column_name].astype("int")
                print("For visualization True is replaced by 1 and False by 0")
            else:
                tmp_dataframe = self.dataframe[column_name]

            histogram(tmp_dataframe)
        pass


def explore_data(dataframe):
    """ Interactive data explorer

    This function should be run in a Jupyter notebook. The user can go through the feature interactively using a slider.

    :param pd.DataFrame dataframe: A pandas dataframe that contains the dataset e.g. train_dataframe.

    """

    explore_object = ExploreData(dataframe)

    logger.info("Starting the interactive data exploring")
    logger.debug(f"The number of columns is {dataframe.shape[1] - 1}")

    interact(
        explore_object.data_explore,
        column_i=widgets.IntSlider(min=0, max=dataframe.shape[1] - 1, step=1, value=0),
    )



import logging
import os
import pickle

import numpy as np
from sklearn.metrics import mean_squared_error, r2_score, accuracy_score, roc_auc_score

logger = logging.getLogger(__name__)
formatting = (
    "%(asctime)s: %(levelname)s: File:%(filename)s Function:%(funcName)s Line:%(lineno)d "
    "message:%(message)s"
)
logging.basicConfig(
    filename=os.path.join(os.path.dirname(os.path.dirname(__file__)), "logs/logs.log"),
    level=logging.INFO,
    format=formatting,
)

metrics_map = {
    "mean_squared_error": mean_squared_error,
    "r2_score": lambda x1, x2: round(100 * r2_score(x1, x2), 2),
    "accuracy_score": lambda x1, x2: round(accuracy_score(x1, np.around(x2)), 4),
    "roc_auc_score": roc_auc_score
}


def load_all_models(save_models_dir: str, model_type: str, model_i: int):
    """ Model loader

    Load saved models from a given type.

    :param str save_models_dir: directory where the model is saved
    :param str model_type: `regression` or `classification`
    :param int model_i: index used to distinguish models of the same type trained on different datasets.
    :return:
    """

    try:
        with open(os.path.join(save_models_dir, f"{model_type}_{model_i}.pkl"), "rb") as save_model:
            model = pickle.load(save_model)
    except Exception as e:
        logger.error(f"Error is: {e}")
        return None

    return model


def evaluate_model(model,
                   xs: list, ys: list, labels: list,
                   metrics: list):
    """ Model evaluator

    This function shows the value of the matrices R2 and MSE for different datasets when evaluating the trained model.

    :param model: An object created by the training package e.g. Scikit Learn.
    :param list xs: Every element is a np.array of the features that are used to predict the target variable.
    :param list ys: Every element is a np.array of the target variable.
    :param list labels: Every element is a string that is used to label every (x,y) pair and refers to their origin.
    :param list metrics: list of metrics used to evaluate model.

    :return:
            metrics_summary (dict): all metrics from `metrics` applied to all (y, y_pred=model(x)) paris.
    """

    metrics_summary = {}

    for x, y, label in zip(xs, ys, labels):

        y_pred = model.predict(x)

        for metric in metrics:
            try:
                value = metrics_map[metric](y, y_pred)
            except Exception as e:
                logger.error(f"Error during metric calculation: {e}")
                print(f"Error during metric calculation: {e}\n"
                      "metric value set to np.nan ")
                value = np.nan

            metrics_summary[f"{metric} {label}"] = value

    logger.info("Evaluate Model process is finished")

    return metrics_summary



from training.utils import split_dataset
import numpy as np
import pandas as pd


# create dataframe
def mock_data(number_of_samples):
    integer_array = np.random.randint(2, size=(number_of_samples, 2))
    for catagories_numbers in range(5, 50, 10):
        integer_array = np.append(
            integer_array, np.random.randint(catagories_numbers, size=(number_of_samples, 2)), axis=1
        )
    integer_columns = [f"int_col_{x}" for x in range(integer_array.shape[1])]

    continuous_array = np.random.randn(number_of_samples, 10)
    continuous_columns = [f"cont_col_{x}" for x in range(continuous_array.shape[1])]

    integer_dataframe = pd.DataFrame(integer_array, columns=integer_columns)
    continuous_dataframe = pd.DataFrame(continuous_array, columns=continuous_columns)

    dataframe = pd.concat([integer_dataframe, continuous_dataframe], axis=1)
    target = (np.sum(continuous_array, axis=1) - 1) / (1 + np.sum(integer_array, axis=1))
    return dataframe, target


train_dataframe, train_target = mock_data(1000)
split_ratios = [0.2]

sub_datasets = split_dataset(train_dataframe.values, train_target, split_ratios)


import unittest
import numpy as np
import pandas as pd
from training.training import model_training


# create dataframe for testing the preprocessing functions:
def mock_data(number_of_samples):
    integer_array = np.random.randint(2, size=(number_of_samples, 2))
    for categories_numbers in range(5, 50, 10):
        integer_array = np.append(
            integer_array, np.random.randint(categories_numbers, size=(number_of_samples, 2)), axis=1
        )
    integer_columns = [f"int_col_{x}" for x in range(integer_array.shape[1])]

    continuous_array = np.random.randn(number_of_samples, 10)
    continuous_columns = [f"cont_col_{x}" for x in range(continuous_array.shape[1])]

    integer_dataframe = pd.DataFrame(integer_array, columns=integer_columns)
    continuous_dataframe = pd.DataFrame(continuous_array, columns=continuous_columns)

    dataframe = pd.concat([integer_dataframe, continuous_dataframe], axis=1)
    target = (np.sum(continuous_array, axis=1) - 1) / (1 + np.sum(integer_array, axis=1))
    return dataframe, target


train_dataframe, train_target = mock_data(1000)
valid_dataframe, valid_target = mock_data(100)
test_dataframe, test_target = mock_data(100)

parameters_linear = {
    "data": {
        "train": {"features": train_dataframe, "target": train_target},
        "valid": {"features": valid_dataframe, "target": valid_target},
        "test": {"features": test_dataframe, "target": test_target},
    },
    "split": {
        "method": "split",  # "method":"kfold"
        "split_ratios": 0.2,  # foldnr:5 , "split_ratios": 0.8 # "split_ratios":(0.7,0.2)
    },
    "model": {"type": "Ridge linear regression",
              "hyperparameters": {"alpha": 1,  # alpha:optimize
                                  },
              },
    "metrics": ["r2_score", "mean_squared_error"],
    "predict": {
        "test": {"features": test_dataframe}
    }
}

parameters_lightgbm = {
    "data": {
        "train": {"features": train_dataframe, "target": train_target},
        "valid": {"features": valid_dataframe, "target": valid_target},
        "test": {"features": test_dataframe, "target": test_target},
    },
    "split": {
        "method": "split",  # "method":"kfold"
        "split_ratios": 0.2,  # foldnr:5 , "split_ratios": 0.8 # "split_ratios":(0.7,0.2)
    },
    "model": {"type": "lightgbm",
              "hyperparameters": dict(objective='regression', metric='root_mean_squared_error', num_leaves=5,
                                      boost_from_average=True,
                                      learning_rate=0.05, bagging_fraction=0.99, feature_fraction=0.99, max_depth=-1,
                                      num_rounds=10000, min_data_in_leaf=10, boosting='dart')
              },
    "metrics": ["r2_score", "mean_squared_error"],
    "predict": {
        "test": {"features": test_dataframe}
    }
}


class MyTestCase(unittest.TestCase):
    def test_something(self):
        model_lists, model_dir = model_training(parameters_linear)
        self.assertEqual(model_lists, ["0"])
        # self.assertEqual(model_dir, 0)


if __name__ == '__main__':
    unittest.main()



import xgboost as xgb
import numpy as np
import pandas as pd
from training.utils import xgboost_problem_type


def xgboost_data_preparation(validation_list: list, dataframe: pd.DataFrame, target: np.array, key: str):
    """ xgboost data preparing for training

    The function transforms the data from a Pandas dataframe format to a xgboost-compatible format.

    :param list validation_list: The list that contains the data the should be used to train and validate the model.
    :param pd.DataFrame dataframe: Pandas dataframe that contains the data which will be transformed to xgboost format.
    :param np.array target: An array that contains the target that should be predict by the xgboost model
    :param str key: A label that is used to name the dataset in the validation_list

    :return:
            - The updated validation_list
    """

    xgboost_data_train = xgb.DMatrix(dataframe, label=target)
    validation_list.append((xgboost_data_train, key))

    return validation_list


def xgboost_regression_train(validation_list: list,
                             hyperparameters: dict,
                             num_round: int = 10):
    """ xgboost trainer

    The function uses the xgboost framework to train the model

    :param list validation_list: The list that contains the data the should be used to train and validate the model.
    :param dict hyperparameters: A dictionary that contains the hyperparameters which the selected training method
            needs to train the model.
    :param int num_round: The number of rounds for boosting

    :return:
            - xgboost model
    """

    param = hyperparameters
    bst = xgb.train(param, validation_list[0][0], num_round, validation_list)
    return bst


def xgboost_data_preparation_to_predict(dataframe: pd.DataFrame):
    """ xgboost data preparing for prediction

    The function transforms the data from a Pandas dataframe format to a xgboost-compatible format

    :param pd.DataFrame dataframe: Pandas dataframe that contains the data which will be transformed to xgboost format.

    :return:
            - The dataset in xgboost-compatible format
    """

    data = xgb.DMatrix(dataframe)

    return data


def training_xgboost_n_split(sub_datasets: dict, hyperparameters: dict, num_round: int = 10):
    """ XGboost training with n-split

    This function trains a model to fit the data using n split cross-validation e.g train, test or train, valid and test

    :param dict sub_datasets:
    :param dict hyperparameters: A dictionary that contains the hyperparameters which the selected training method
            need to train the model.
    :param int num_round: The number of rounds for boosting

    :return:
            - model: xgboost model.
            - problem_to_solve: string that defines the problem to solve: regression or classification.
            - validation_list: The list that contains the data the should be used to train and validate the model.
    """

    # Find out what is the type of the problem that should be solved based on the defined objective
    problem_to_solve = xgboost_problem_type(hyperparameters)

    # train data transformation
    validation_list = []
    data_i = "train"
    dataframe = pd.DataFrame(sub_datasets["train"])
    target = sub_datasets["target"]
    validation_list = xgboost_data_preparation(validation_list, dataframe, target, data_i)

    sub_datasets["train"] = validation_list[-1][0]

    # transformation of the reset of the datasets
    for index_i in range(len(list(sub_datasets.keys()))):
        try:
            data_i = f"dataset_{index_i}"
            print(data_i)
            dataframe = pd.DataFrame(sub_datasets[f"features_{index_i}"])
            target = sub_datasets[f"target_{index_i}"]
            validation_list = xgboost_data_preparation(validation_list, dataframe, target, data_i)
            sub_datasets[f"features_{index_i}"] = validation_list[-1][0]
        except:
            print("All dataset are considered")
            break

    model = xgboost_regression_train(validation_list, hyperparameters, num_round=num_round)

    return model, problem_to_solve, validation_list


def training_xgboost_kfold(train_array, target, train: list, test: list, hyperparameters: dict, num_round: int = 10):
    """ XGboost training with kfold

    This function trains a model to fit the data using K-Fold cross-validation.

    :param np.array train_array: The values of the target that will be split into K-Folds and used to train the model to
            predict the target
    :param np.array target: The values of the target that will be split into K-Folds and used to train the model.
    :param list train: A list of integers that define the training dataset
    :param list test: A list of integers that define the testing dataset
    :param dict hyperparameters: A dictionary that contains the hyperparameters which the selected training method
            needs to train the model.
    :param int num_round: The number of rounds for boosting

    :return:
            - kfold_model: xgboost model
            - problem_to_solve: string that defines the problem to solve: regression or classification.
            - validation_list: The list that contains the data the should be used to train and validate the model.
    """

    # train data transformation
    validation_list = []
    data_i = "train"
    dataframe = pd.DataFrame(train_array[train])
    xgboost_target = target[train]
    validation_list = xgboost_data_preparation(validation_list, dataframe, xgboost_target, data_i)

    # test data transformation
    data_i = "test"
    dataframe = pd.DataFrame(train_array[test])
    xgboost_target = target[test]
    validation_list = xgboost_data_preparation(validation_list, dataframe, xgboost_target, data_i)

    kfold_model = xgboost_regression_train(validation_list, hyperparameters, num_round=num_round)

    return kfold_model, validation_list


def get_num_round(hyperparameters) -> int:
    """ num_round getter

    Get the value of num_round that will be used to train the xgboost model

    :param hyperparameters: A dictionary that contains the hyperparameters which the selected training method
            needs to train the model.
    :return:
            - num_round: The number of rounds for boosting
    """

    try:
        num_round = hyperparameters["num_round"]
    except Exception as e:
        print(f"The num_round is not defined. The default value is num_round = 10. Error: {e}")
        num_round = 10
    return num_round


def xgboost_data_preparation_for_evaluation(data: dict):
    """ Date preparation for evaluation

    Prepare the data in a form that could be used for model evaluation.

    :param data:
    :return:
    """

    for data_i in data.keys():
        dataframe = data[data_i]["features"]
        dataframe.columns = [x for x in range(len(list(dataframe.columns)))]
        target = data[data_i]["target"]

        data[data_i]["features"] = xgb.DMatrix(dataframe, label=target)

    return data
